"""The Pursuit-Evasion Grid World Environment.

An adversarial 2D grid world problem involving two agents, a evader and
a pursuer. The evader's goal is to reach a goal location, on the other side
of the grid, while the goal of the pursuer is to spot the evader
before it reaches it's goal. The evader is considered caught if it is
observed by the pursuer, or occupies the same location. The evader and
pursuer have knowledge of each others starting locations. However, only the
evader has knowledge of it's goal location. The pursuer only knowns that
the goal location is somewhere on the opposite side of the grid to the
evaders start location.

This environment requires each agent to reason about the which path the
other agent will take through the dense grid environment.

References
----------
- [This Pursuit-Evasion implementation is directly inspired by the problem] Seaman, Iris
  Rubi, Jan-Willem van de Meent, and David Wingate. 2018. “Nested Reasoning About
  Autonomous Agents Using Probabilistic Programs.” ArXiv Preprint ArXiv:1812.01569.
- Schwartz, Jonathon, Ruijia Zhou, and Hanna Kurniawati. "Online Planning for
  Interactive-POMDPs using Nested Monte Carlo Tree Search." In 2022 IEEE/RSJ
  International Conference on Intelligent Robots and Systems (IROS), pp. 8770-8777.
  IEEE, 2022.

"""
from collections import deque
from typing import (
    Any,
    Deque,
    Dict,
    List,
    NamedTuple,
    Optional,
    Set,
    SupportsFloat,
    Tuple,
    Union,
)

from gymnasium import spaces

import posggym.envs.grid_world.render as render_lib
import posggym.model as M
from posggym.core import DefaultEnv
from posggym.envs.grid_world.core import Coord, Direction, Grid
from posggym.utils import seeding


# State = (e_coord, e_dir, p_coord, p_dir, e_0_coord, p_0_coord, e_goal_coord)
INITIAL_DIR = Direction.NORTH


class PEState(NamedTuple):
    """Environment state in Pursuit Evastion problem."""

    evader_coord: Coord
    evader_dir: Direction
    pursuer_coord: Coord
    pursuer_dir: Direction
    evader_start_coord: Coord
    pursuer_start_coord: Coord
    evader_goal_coord: Coord
    min_goal_dist: int  # for evader


# Action = Direction of movement
# 0 = Forward, 1 = Backward, 2 = Left, 3 = Right
PEAction = int
ACTION_TO_DIR = [
    # Forward
    [Direction.NORTH, Direction.EAST, Direction.SOUTH, Direction.WEST],
    # Backward
    [Direction.SOUTH, Direction.WEST, Direction.NORTH, Direction.EAST],
    # Left
    [Direction.WEST, Direction.NORTH, Direction.EAST, Direction.SOUTH],
    # Right
    [Direction.EAST, Direction.SOUTH, Direction.WEST, Direction.NORTH],
]

# E Obs = Tuple[WallObs, seen, heard, e_0_coord, p_0_coord, goal_coord]
#       = Tuple[Tuple[int, int, int, int, int, int], Coord, Coord, Coord]
# P Obs = Tuple[WallObs, seen, heard, e_0_coord, p_0_coord, blank_coord]
#       = Tuple[Tuple[int, int, int, int, int, int], Coord, Coord, Coord]
# Note, we use blank_coord for P Obs so Obs spaces are identical between the
# two agents. The blank_coord is always (0, 0).
PEEvaderObs = Tuple[Tuple[int, ...], Coord, Coord, Coord]
PEPursuerObs = Tuple[Tuple[int, ...], Coord, Coord, Coord]
PEObs = Union[PEEvaderObs, PEPursuerObs]


class PursuitEvasionEnv(DefaultEnv):
    """The Pursuit-Evasion Grid World Environment.

    An adversarial 2D grid world problem involving two agents, a evader and
    a pursuer. The evader's goal is to reach a goal location, on the other side
    of the grid, while the goal of the pursuer is to spot the evader
    before it reaches it's goal. The evader is considered caught if it is
    observed by the pursuer, or occupies the same location. The evader and
    pursuer have knowledge of each others starting locations. However, only the
    evader has knowledge of it's goal location. The pursuer only knowns that
    the goal location is somewhere on the opposite side of the grid to the
    evaders start location.

    This environment requires each agent to reason about the which path the
    other agent will take through the dense grid environment.

    Agents
    ------
    Evader=0
    Pursuer=1

    State
    -----
    Each state is made up of:

    0. the (x, y) coordinate of the evader
    1. the direction the evader is facing
    2. the (x, y) coordinate of the pursuer
    3. the direction the pursuer is facing
    4. the (x, y) coordinate of the evader
    5. the (x, y) coordinate of the evader's start location
    6. the (x, y) coordinate of the pursuer's start location

    Actions
    -------
    Each agent has 4 actions corresponding to moving in the 4 available
    directions w.r.t the direction the agent is currently faction
    (FORWARD=0, BACKWARDS=1, LEFT=2, RIGHT=3).

    Observation
    -----------
    Each agent observes:

    1. whether there is a wall or not in the adjacent cells in the four
       cardinal directions,
    2. whether they see the other agent in a cone in front of them. The cone
       projects forward up to 'max_obs_distance' (default=12) cells in front of
       the agent.
    3. whether they hear the other agent (whether the other agent is within
       distance 2 from the agent in any direction),
    4. the (x, y) coordinate of the evader's start location,
    5. the (x, y) coordinate of the pursuer's start location,
    6. Evader: the (x, y) coordinate of the evader's goal location.
       Pursuer: blank coordinate = (0, 0).

    Note, the goal and start coordinate observations do not change during a
    single episode, but they do change between episodes.

    Reward
    ------
    The environment is zero-sum with the pursuer recieving the negative of the
    evader reward. Additionally, rewards are by default normalized so that
    returns are bounded between -1 and 1 (this can be disabled by the
    `normalize_reward` parameter).

    The evader recieves a reward of 1 for reaching it's goal location and a
    reward of -1 if it gets captured. Additionally, the evader recieves a small
    reward of 0.01 each time it's minumum distance to it's goal along the
    shortest path decreases. This is to make it so the environment is no
    longer sparesely rewarded and to help with exploration and learning (it
    can be disabled by the `use_progress_reward` parameter.)

    Transition Dynamics
    -------------------
    By default actions are deterministic and an episode ends when either the
    evader is caught, the evader reaches a goal, or the step limit is reached.

    The environment can also be run in stochastic mode by changing the
    `action_probs` parameter at initialization. This controls the probability
    the agent will move in the desired direction each step, otherwise moving
    randomly in one of the other 3 possible directions.

    References
    ----------
    - [This Pursuit-Evasion implementation is directly inspired by the problem] Seaman,
      Iris Rubi, Jan-Willem van de Meent, and David Wingate. 2018. “Nested Reasoning
      About Autonomous Agents Using Probabilistic Programs.”
      ArXiv Preprint ArXiv:1812.01569.
    - Schwartz, Jonathon, Ruijia Zhou, and Hanna Kurniawati. "Online Planning for
      Interactive-POMDPs using Nested Monte Carlo Tree Search." In 2022 IEEE/RSJ
      International Conference on Intelligent Robots and Systems (IROS), pp. 8770-8777.
      IEEE, 2022.

    """

    metadata = {
        "render_modes": ["human", "ansi", "rgb_array", "rgb_array_dict"],
        "render_fps": 4,
    }

    def __init__(
        self,
        grid_name: str,
        action_probs: Union[float, Tuple[float, float]] = 1.0,
        max_obs_distance: int = 12,
        normalize_reward: bool = True,
        use_progress_reward: bool = True,
        render_mode: Optional[str] = None,
        **kwargs,
    ):
        model = PursuitEvasionModel(
            grid_name,
            action_probs=action_probs,
            max_obs_distance=max_obs_distance,
            normalize_reward=normalize_reward,
            use_progress_reward=use_progress_reward,
            **kwargs,
        )
        super().__init__(
            model,
            render_mode=render_mode,
        )
        self._max_obs_distance = max_obs_distance
        grid = model.grid
        fov_width = grid.get_max_fov_width(model.FOV_EXPANSION_INCR, max_obs_distance)

        self._obs_dims = (
            min(max_obs_distance, max(grid.width, grid.height)),
            0,
            fov_width // 2,
        )
        self._viewer = None
        self._renderer: Optional[render_lib.GWRenderer] = None

    def reset(
        self, *, seed: Optional[int] = None, options: Optional[Dict[str, Any]] = None
    ) -> Tuple[Optional[Dict[M.AgentID, M.ObsType]], Dict[M.AgentID, Dict]]:
        # reset renderer since goal location can change between episodes
        self._renderer = None
        return super().reset(seed=seed, options=options)

    def render(self):
        evader_coord = self._state[0]
        pursuer_coord = self._state[2]
        goal_coord = self._state[6]
        grid = self.model.grid  # type: ignore

        if self.render_mode == "ansi":
            grid_str = grid.get_ascii_repr(goal_coord, evader_coord, pursuer_coord)
            output = [
                f"Step: {self._step_num}",
                grid_str,
            ]
            if self._last_actions is not None:
                action_str = ", ".join(
                    [str(Direction(a)) for a in self._last_actions.values()]
                )
                output.insert(1, f"Actions: <{action_str}>")
                output.append(f"Rewards: <{self._last_rewards}>")
            return "\n".join(output) + "\n"

        if self.render_mode == "human" and self._viewer is None:
            # pylint: disable=[import-outside-toplevel]
            from posggym.envs.grid_world import viewer

            self._viewer = viewer.GWViewer(
                "Pursuit-Evasion Env",
                (min(grid.height, 8), min(grid.width, 8)),
                num_agent_displays=len(self.possible_agents),
            )
            self._viewer.show(block=False)

        if self._renderer is None:
            self._renderer = render_lib.GWRenderer(
                len(self.possible_agents), grid, [], render_blocks=True
            )

        agent_obs_coords = tuple(
            list(
                grid.get_fov(
                    self._state[2 * i],
                    self._state[2 * i + 1],
                    self.model.FOV_EXPANSION_INCR,
                    self._max_obs_distance,
                )
            )
            for i in range(len(self.possible_agents))
        )
        agent_coords = (evader_coord, pursuer_coord)
        agent_dirs = (self._state[1], self._state[3])

        other_objs = [
            render_lib.GWObject(goal_coord, "green", render_lib.Shape.RECTANGLE)
        ]

        env_img = self._renderer.render(
            agent_coords,
            agent_obs_coords,
            agent_dirs=agent_dirs,
            other_objs=other_objs,
            agent_colors=None,
        )

        agent_obs_imgs = self._renderer.render_all_agent_obs(
            env_img,
            agent_coords,
            agent_dirs,
            agent_obs_dims=self._obs_dims,
            out_of_bounds_obj=render_lib.GWObject(
                (0, 0), "grey", render_lib.Shape.RECTANGLE
            ),
            agent_obs_coords=agent_obs_coords,
        )

        if self.render_mode == "human":
            self._viewer.update_img(env_img, agent_idx=None)
            for i, obs_img in enumerate(agent_obs_imgs):
                self._viewer.update_img(obs_img, agent_idx=i)
            self._viewer.display_img()
        elif self.render_mode == "rgb_array":
            return env_img
        else:
            # rgb_array_dict
            return dict(enumerate(agent_obs_imgs))

    def close(self) -> None:
        if self._viewer is not None:
            self._viewer.close()  # type: ignore
            self._viewer = None


class PursuitEvasionModel(M.POSGModel[PEState, PEObs, PEAction]):
    """Discrete Pursuit-Evasion Model."""

    NUM_AGENTS = 2

    EVADER_IDX = 0
    PURSUER_IDX = 1

    # Evader-centric rewards
    R_PROGRESS = 0.01  # Reward making progress toward goal
    R_EVADER_ACTION = -0.01  # Reward each step for evader
    R_PURSUER_ACTION = -0.01  # Reward each step for pursuer
    R_CAPTURE = -1.0  # Reward for being captured
    R_EVASION = 1.0  # Reward for reaching goal

    FOV_EXPANSION_INCR = 3
    HEARING_DIST = 2

    def __init__(
        self,
        grid_name: str,
        action_probs: Union[float, Tuple[float, float]] = 1.0,
        max_obs_distance: int = 12,
        normalize_reward: bool = True,
        use_progress_reward: bool = True,
        **kwargs,
    ):
        self._grid_name = grid_name
        self.grid = load_grid(grid_name)
        if isinstance(action_probs, float):
            action_probs = (action_probs, action_probs)
        self._action_probs = action_probs
        self._max_obs_distance = max_obs_distance
        self._normalize_reward = normalize_reward
        self._use_progress_reward = use_progress_reward

        self._max_sp_distance = self.grid.get_max_shortest_path_distance()
        self._max_raw_return = self.R_EVASION
        if self._use_progress_reward:
            self._max_raw_return += (self._max_sp_distance + 1) * self.R_PROGRESS
        self._min_raw_return = -self._max_raw_return

        def _coord_space():
            return spaces.Tuple(
                (spaces.Discrete(self.grid.width), spaces.Discrete(self.grid.height))
            )

        self.possible_agents = tuple(range(self.NUM_AGENTS))
        # s = Tuple[Coord, Direction, Coord, Direction, Coord, Coord, Coord, int]
        # e_coord, e_dir, p_coord, p_dir, e_start, p_start, e_goal, max_sp
        self.state_space = spaces.Tuple(
            (
                _coord_space(),
                spaces.Discrete(len(Direction)),
                _coord_space(),
                spaces.Discrete(len(Direction)),
                _coord_space(),
                _coord_space(),
                _coord_space(),
                spaces.Discrete(self._max_sp_distance + 1),
            )
        )
        self.action_spaces = {
            i: spaces.Discrete(len(Direction)) for i in self.possible_agents
        }
        # o = Tuple[Tuple[WallObs, seen , heard], Coord, Coord, Coord]
        # Wall obs, seen, heard, e_start, p_start, e_goal/blank
        self.observation_spaces = {
            i: spaces.Tuple(
                (
                    spaces.Tuple(tuple(spaces.Discrete(2) for _ in range(6))),
                    _coord_space(),
                    _coord_space(),
                    _coord_space(),
                )
            )
            for i in self.possible_agents
        }
        self.observation_first = True
        self.is_symmetric = False

    @property
    def reward_ranges(self) -> Dict[M.AgentID, Tuple[SupportsFloat, SupportsFloat]]:
        max_reward = self.R_CAPTURE
        if self._use_progress_reward:
            max_reward += self.R_PROGRESS
        if self._normalize_reward:
            max_reward = self._get_normalized_reward(max_reward)
        return {i: (-max_reward, max_reward) for i in self.possible_agents}

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: PEState) -> List[M.AgentID]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> PEState:
        return self._sample_initial_state(None, None, None)

    def sample_agent_initial_state(self, agent_id: M.AgentID, obs: PEObs) -> PEState:
        if agent_id == self.EVADER_IDX:
            return self._sample_initial_state(
                evader_coord=obs[1], pursuer_coord=obs[2], goal_coord=obs[3]
            )
        return self._sample_initial_state(
            evader_coord=obs[1], pursuer_coord=obs[2], goal_coord=None
        )

    def _sample_initial_state(
        self,
        evader_coord: Optional[Coord],
        pursuer_coord: Optional[Coord],
        goal_coord: Optional[Coord],
    ) -> PEState:
        if evader_coord is None:
            evader_coord = self.rng.choice(self.grid.evader_start_coords)
        if pursuer_coord is None:
            pursuer_coord = self.rng.choice(self.grid.pursuer_start_coords)
        if goal_coord is None:
            goal_coord = self.rng.choice(self.grid.get_goal_coords(evader_coord))
        return PEState(
            evader_coord,
            INITIAL_DIR,
            pursuer_coord,
            INITIAL_DIR,
            evader_coord,
            pursuer_coord,
            goal_coord,
            self.grid.get_shortest_path_distance(evader_coord, goal_coord),
        )

    def sample_initial_obs(self, state: PEState) -> Dict[M.AgentID, PEObs]:
        return self._get_obs(state)[0]

    def step(
        self, state: PEState, actions: Dict[M.AgentID, PEAction]
    ) -> M.JointTimestep[PEState, PEObs]:
        assert all(0 <= a_i < len(Direction) for a_i in actions.values())
        next_state = self._get_next_state(state, actions)
        obs, evader_detected = self._get_obs(next_state)
        rewards = self._get_reward(state, next_state, evader_detected)
        all_done = self._is_done(next_state)
        terminated = {i: all_done for i in self.possible_agents}
        truncated = {i: False for i in self.possible_agents}
        info: Dict[M.AgentID, Dict] = {i: {} for i in self.possible_agents}
        if all_done:
            for i, outcome in self._get_outcome(next_state).items():
                info[i]["outcome"] = outcome
        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _get_next_state(
        self, state: PEState, actions: Dict[M.AgentID, PEAction]
    ) -> PEState:
        evader_a = actions[self.EVADER_IDX]
        pursuer_a = actions[self.PURSUER_IDX]

        if (
            self._action_probs[self.EVADER_IDX] < 1.0
            and self.rng.random() > self._action_probs[self.EVADER_IDX]
        ):
            other_as = [a for a in range(len(Direction)) if a != evader_a]
            evader_a = self.rng.choice(other_as)

        if (
            self._action_probs[self.PURSUER_IDX]
            and self.rng.random() > self._action_probs[self.PURSUER_IDX]
        ):
            other_as = [a for a in range(len(Direction)) if a != pursuer_a]
            pursuer_a = self.rng.choice(other_as)

        pursuer_next_dir = Direction(ACTION_TO_DIR[pursuer_a][state.pursuer_dir])
        pursuer_next_coord = self.grid.get_next_coord(
            state.pursuer_coord, pursuer_next_dir, ignore_blocks=False
        )

        evader_next_coord = state.evader_coord
        evader_next_dir = Direction(ACTION_TO_DIR[evader_a][state.evader_dir])
        if pursuer_next_coord != state.evader_coord:
            evader_next_coord = self.grid.get_next_coord(
                state.evader_coord, evader_next_dir, ignore_blocks=False
            )

        min_sp_distance = min(
            state.min_goal_dist,
            self.grid.get_shortest_path_distance(
                evader_next_coord, state.evader_goal_coord
            ),
        )

        return PEState(
            evader_next_coord,
            evader_next_dir,
            pursuer_next_coord,
            pursuer_next_dir,
            state.evader_start_coord,
            state.pursuer_start_coord,
            state.evader_goal_coord,
            min_sp_distance,
        )

    def _get_obs(self, state: PEState) -> Tuple[Dict[M.AgentID, PEObs], bool]:
        walls, seen, heard = self._get_agent_obs(
            state.evader_coord, state.evader_dir, state.pursuer_coord
        )
        evader_obs = (
            (*walls, seen, heard),
            state.evader_start_coord,
            state.pursuer_start_coord,
            state.evader_goal_coord,
        )

        walls, seen, heard = self._get_agent_obs(
            state.pursuer_coord, state.pursuer_dir, state.evader_coord
        )
        pursuer_obs = (
            (*walls, seen, heard),
            state.evader_start_coord,
            state.pursuer_start_coord,
            # pursuer doesn't observe evader goal coord
            (0, 0),
        )
        return {self.EVADER_IDX: evader_obs, self.PURSUER_IDX: pursuer_obs}, bool(seen)

    def _get_agent_obs(
        self, agent_coord: Coord, agent_dir: Direction, opp_coord: Coord
    ) -> Tuple[Tuple[int, int, int, int], int, int]:
        adj_coords = self.grid.get_neighbours(
            agent_coord, ignore_blocks=True, include_out_of_bounds=True
        )
        walls: Tuple[int, int, int, int] = tuple(  # type: ignore
            int(
                not self.grid.coord_in_bounds(agent_coord)
                or agent_coord in self.grid.block_coords
            )
            for coord in adj_coords
        )
        seen = self._get_opponent_seen(agent_coord, agent_dir, opp_coord)
        dist = self.grid.manhattan_dist(agent_coord, opp_coord)
        heard = dist <= self.HEARING_DIST
        return walls, int(seen), int(heard)

    def _get_opponent_seen(
        self, ego_coord: Coord, ego_dir: Direction, opp_coord: Coord
    ) -> bool:
        fov = self.grid.get_fov(
            ego_coord, ego_dir, self.FOV_EXPANSION_INCR, self._max_obs_distance
        )
        return opp_coord in fov

    def _get_reward(
        self, state: PEState, next_state: PEState, evader_seen: bool
    ) -> Dict[M.AgentID, SupportsFloat]:
        evader_coord = next_state.evader_coord
        pursuer_coord = next_state.pursuer_coord
        evader_goal_coord = next_state.evader_goal_coord

        evader_reward = 0.0
        if self._use_progress_reward and next_state.min_goal_dist < state.min_goal_dist:
            evader_reward += self.R_PROGRESS

        if evader_coord == pursuer_coord or evader_seen:
            evader_reward += self.R_CAPTURE
        elif evader_coord == evader_goal_coord:
            evader_reward += self.R_EVASION

        if self._normalize_reward:
            evader_reward = self._get_normalized_reward(evader_reward)

        return {self.EVADER_IDX: evader_reward, self.PURSUER_IDX: -evader_reward}

    def _is_done(self, state: PEState) -> bool:
        evader_coord, pursuer_coord = state.evader_coord, state.pursuer_coord
        pursuer_dir = state.pursuer_dir
        return (
            evader_coord == pursuer_coord
            or evader_coord == state.evader_goal_coord
            or self._get_opponent_seen(pursuer_coord, pursuer_dir, evader_coord)
        )

    def _get_outcome(self, state: PEState) -> Dict[M.AgentID, M.Outcome]:
        # Assuming this method is called on final timestep
        evader_coord, pursuer_coord = state.evader_coord, state.pursuer_coord
        evader_goal_coord = state.evader_goal_coord
        pursuer_dir = state.pursuer_dir
        # check this first before relatively expensive detection check
        if evader_coord == pursuer_coord:
            return {self.EVADER_IDX: M.Outcome.LOSS, self.PURSUER_IDX: M.Outcome.WIN}
        if evader_coord == evader_goal_coord:
            return {self.EVADER_IDX: M.Outcome.WIN, self.PURSUER_IDX: M.Outcome.LOSS}
        if self._get_opponent_seen(pursuer_coord, pursuer_dir, evader_coord):
            return {self.EVADER_IDX: M.Outcome.LOSS, self.PURSUER_IDX: M.Outcome.WIN}
        return {self.EVADER_IDX: M.Outcome.DRAW, self.PURSUER_IDX: M.Outcome.DRAW}

    def _get_normalized_reward(self, reward: float) -> float:
        """Normalize reward in [-1, 1] interval."""
        diff = self._max_raw_return - self._min_raw_return
        return 2 * (reward - self._min_raw_return) / diff - 1


class PEGrid(Grid):
    """A grid for the Pursuit Evasion Problem."""

    def __init__(
        self,
        grid_width: int,
        grid_height: int,
        block_coords: Set[Coord],
        goal_coords_map: Dict[Coord, List[Coord]],
        evader_start_coords: List[Coord],
        pursuer_start_coords: List[Coord],
    ):
        super().__init__(grid_width, grid_height, block_coords)
        self._goal_coords_map = goal_coords_map
        self.evader_start_coords = evader_start_coords
        self.pursuer_start_coords = pursuer_start_coords

        self.shortest_paths = self.get_all_shortest_paths(set(evader_start_coords))

    @property
    def all_goal_coords(self) -> List[Coord]:
        """The list of all evader goal locations."""
        all_locs = set()
        for v in self._goal_coords_map.values():
            all_locs.update(v)
        return list(all_locs)

    def get_goal_coords(self, evader_start_coord: Coord) -> List[Coord]:
        """Get list of possible evader goal coords for given start coords."""
        return self._goal_coords_map[evader_start_coord]

    def get_shortest_path_distance(self, coord: Coord, dest: Coord) -> int:
        """Get the shortest path distance from coord to destination."""
        return int(self.shortest_paths[dest][coord])

    def get_max_shortest_path_distance(self) -> int:
        """Get max shortest path distance between any start and goal coords."""
        max_dist = 0
        for start_coord, dest_coords in self._goal_coords_map.items():
            max_dist = max(
                max_dist,
                int(
                    max(
                        [
                            self.get_shortest_path_distance(start_coord, dest)
                            for dest in dest_coords
                        ]
                    )
                ),
            )
        return max_dist

    def get_ascii_repr(
        self,
        goal_coord: Union[None, Coord, List[Coord]],
        evader_coord: Union[None, Coord, List[Coord]],
        pursuer_coord: Union[None, Coord, List[Coord]],
    ) -> str:
        """Get ascii repr of grid."""
        if goal_coord is None:
            goal_coords = set()
        elif isinstance(goal_coord, list):
            goal_coords = set(goal_coord)
        else:
            goal_coords = set([goal_coord])

        grid_repr = []
        for row in range(self.height):
            row_repr = []
            for col in range(self.width):
                coord = (col, row)
                if coord in goal_coords:
                    row_repr.append("G")
                elif coord in self.block_coords:
                    row_repr.append("#")
                else:
                    row_repr.append(".")
            grid_repr.append(row_repr)

        if evader_coord is None:
            evader_coord = []
        elif not isinstance(evader_coord, List):
            evader_coord = [evader_coord]

        for coord in evader_coord:
            grid_repr[coord[1]][coord[0]] = "R"

        if pursuer_coord is None:
            pursuer_coord = []
        elif not isinstance(pursuer_coord, List):
            pursuer_coord = [pursuer_coord]

        for coord in pursuer_coord:
            if coord in evader_coord:
                grid_repr[coord[1]][coord[0]] = "X"
            else:
                grid_repr[coord[1]][coord[0]] = "C"

        return "\n".join(list(list((" ".join(r) for r in grid_repr))))

    def get_init_ascii_repr(self) -> str:
        """Get ascii repr of initial grid."""
        return self.get_ascii_repr(
            self.all_goal_coords, self.evader_start_coords, self.pursuer_start_coords
        )

    def get_fov(
        self,
        origin: Coord,
        direction: Direction,
        widening_increment: int,
        max_depth: int,
    ) -> Set[Coord]:
        """Get the Field of vision from origin looking in given direction.

        Uses BFS starting from origin and expanding in the direction, while
        accounting for obstacles blocking the field of view,
        to get the field of vision.
        """
        assert widening_increment > 0
        assert max_depth > 0
        fov = set([origin])

        frontier_queue: Deque[Coord] = deque([origin])
        visited = set([origin])

        while len(frontier_queue):
            coord = frontier_queue.pop()

            for next_coord in self._get_fov_successors(
                origin, direction, coord, widening_increment, max_depth
            ):
                if next_coord not in visited:
                    visited.add(next_coord)
                    frontier_queue.append(next_coord)
                    fov.add(next_coord)
        return fov

    def _get_fov_successors(
        self,
        origin: Coord,
        direction: Direction,
        coord: Coord,
        widening_increment: int,
        max_depth: int,
    ) -> List[Coord]:
        if direction in [Direction.NORTH, Direction.SOUTH]:
            depth = abs(origin[1] - coord[1])
        else:
            depth = abs(origin[0] - coord[0])

        if depth >= max_depth:
            return []

        successors = []
        forward_successor = self._get_fov_successor(coord, direction)
        if forward_successor is None:
            return []
        else:
            successors.append(forward_successor)

        # Expands FOV at depth 1 and then every widening_increment depth
        if depth != 1 and depth % widening_increment != 0:
            # Don't expand sideways
            return successors

        side_coords_list: List[Coord] = []

        if direction in [Direction.NORTH, Direction.SOUTH]:
            if 0 < coord[0] <= origin[0]:
                side_coords_list.append((coord[0] - 1, coord[1]))
            if origin[0] <= coord[0] < self.width - 1:
                side_coords_list.append((coord[0] + 1, coord[1]))
        else:
            if 0 < coord[1] <= origin[1]:
                side_coords_list.append((coord[0], coord[1] - 1))
            elif origin[1] <= coord[1] < self.height - 1:
                side_coords_list.append((coord[0], coord[1] + 1))

        side_successor: Optional[Coord] = None
        for side_coord in side_coords_list:
            if side_coord in self.block_coords:
                continue

            side_successor = self._get_fov_successor(side_coord, direction)
            if side_successor is not None:
                successors.append(side_successor)

        return successors

    def _get_fov_successor(self, coord: Coord, direction: Direction) -> Optional[Coord]:
        new_coord = self.get_next_coord(coord, direction, ignore_blocks=False)
        if new_coord == coord:
            # move in given direction is blocked or out-of-bounds
            return None
        return new_coord

    def get_max_fov_width(self, widening_increment: int, max_depth: int) -> int:
        """Get the maximum width of field of vision."""
        max_width = 1
        for d in range(max_depth):
            if d == 1 or d % widening_increment == 0:
                max_width += 2
        return min(max_depth, min(self.width, self.height))


def get_5x5_grid() -> PEGrid:
    """Generate the 5-by-5 PE grid layout.

    - 0, 1, 2, 3 are possible evader start and goal locations
    - 5 are possible pursuer start locations

    The evader start and goal locations are always on opposite sides of the
    map.

    """
    ascii_map = " 9  8" " # # " " #5  " " # # " "0  1 "

    return _convert_map_to_grid(
        ascii_map,
        5,
        5,
        pursuer_start_symbols=set(["5"]),
        evader_start_symbols=set(["0", "1", "8", "9"]),
        evader_goal_symbol_map={
            "0": ["8", "9"],
            "1": ["8", "9"],
            "8": ["0", "1"],
            "9": ["0", "1"],
        },
    )


def get_8x8_grid() -> PEGrid:
    """Generate the 8-by-8 PE grid layout.

    This is an approximate discrete version of the map layout from:
    Seaman et al 2018, 'Nested Reasoning About Autonomous Agents Using
    Probabilistic Programs'

    - 0, 1, 2, 7, 8, 9 are possible evader start and goal locations
    - 5, 6 are possible pursuer start locations

    The evader start and goal locations are always on opposite sides of the
    map.
    """
    ascii_map = (
        "  9 #8 #"
        "# #    #"
        "# ## # 7"
        "  6   # "
        "# #5# # "
        "   #    "
        "#    #2 "
        "0 #1### "
    )

    return _convert_map_to_grid(ascii_map, 8, 8)


def get_16x16_grid() -> PEGrid:
    """Generate the 16-by-16 PE grid layout.

    This is an approximate discrete version of the map layout from:
    Seaman et al 2018, 'Nested Reasoning About Autonomous Agents Using
    Probabilistic Programs'

    - 0, 1, 2, 7, 8, 9 are possible evader start and goal locations
    - 5, 6 are possible pursuer start locations

    The evader start and goal locations are always on opposite sides of the
    map.

    """
    ascii_map = (
        "  ## ####### ###"
        "    9##    8 ###"
        "##       # # ###"
        "## ## ##      ##"
        "## ## ##   ## ##"
        "#  #  ##   ## 7#"
        "# ## ###   #   #"
        "  #  6       # #"
        "#   ## ##  ##  #"
        "##   #5##  #   #"
        "## #   #     # #"
        "   # #   ##    #"
        " # # ##  ##   ##"
        "0#       #     #"
        "   ### #   ##2  "
        "###    1 #####  "
    )
    return _convert_map_to_grid(ascii_map, 16, 16)


def get_32x32_grid() -> PEGrid:
    """Generate the 32-by-32 PE grid layout.

    This is an approximate discrete version of the map layout from:
    Seaman et al 2018, 'Nested Reasoning About Autonomous Agents Using
    Probabilistic Programs'

    - 0, 1, 2, 7, 8, 9 are possible evader start and goal locations
    - 5, 6 are possible pursuer start locations

    The evader start and goal locations are always on opposite sides of the
    map.

    """
    ascii_map = (
        "       #  ########### #         "
        "   #####  ########### #  #######"
        "      #   ######      8  #######"
        " #       9#               ######"
        " ##              #### ##  ######"
        " ##   #####  ##  #### ##   #####"
        " ##   #####  ##  ##        #####"
        "      ##### ####      ###  #####"
        " ### ###### ####      ###   ####"
        " ### #####  ####      ####  ####"
        " ##  ##### #####      ##### 7###"
        " ##  ####  #####       ####  ###"
        " #  #####  #####       ####  ###"
        " #  ##### ######       #     ###"
        " #       6               ##  ###"
        "    #       #   ##      ####  ##"
        "     #   ####  ###     ####   ##"
        "#### #   ####  ###    ####    ##"
        "#### #   ### 5####   ####     ##"
        "####  #      ####     ##  #   ##"
        "##### #      ####        ##   ##"
        "#                          ##  #"
        "         ###        ##    2 #  #"
        "  ###### ###      #### ## #    #"
        "########  ##      ####    #  #  "
        "########  ##       ####     ####"
        "###          ##   1##        ###"
        "          #  ####       #       "
        "   0  ###### ######   ####      "
        "  ########## #        #####     "
        "##########      ############    "
        "#                               "
    )
    return _convert_map_to_grid(ascii_map, 32, 32)


def _loc_to_coord(loc: int, grid_width: int) -> Coord:
    return (loc % grid_width, loc // grid_width)


def _convert_map_to_grid(
    ascii_map: str,
    height: int,
    width: int,
    block_symbol: str = "#",
    pursuer_start_symbols: Optional[Set[str]] = None,
    evader_start_symbols: Optional[Set[str]] = None,
    evader_goal_symbol_map: Optional[Dict] = None,
) -> PEGrid:
    assert len(ascii_map) == height * width

    if pursuer_start_symbols is None:
        pursuer_start_symbols = set(["3", "4", "5", "6"])
    if evader_start_symbols is None:
        evader_start_symbols = set(["0", "1", "2", "7", "8", "9"])
    if evader_goal_symbol_map is None:
        evader_goal_symbol_map = {
            "0": ["7", "8", "9"],
            "1": ["7", "8", "9"],
            "2": ["8", "9"],
            "7": ["0", "1"],
            "8": ["0", "1", "2"],
            "9": ["0", "1", "2"],
        }

    block_coords = set()
    evader_start_coords = []
    pursuer_start_coords = []
    evader_symbol_coord_map = {}

    for loc, symbol in enumerate(ascii_map):
        coord = _loc_to_coord(loc, width)
        if symbol == block_symbol:
            block_coords.add(coord)
        elif symbol in pursuer_start_symbols:
            pursuer_start_coords.append(coord)
        elif symbol in evader_start_symbols:
            evader_start_coords.append(coord)
            evader_symbol_coord_map[symbol] = coord

    evader_goal_coords_map = {}
    for start_symbol, goal_symbols in evader_goal_symbol_map.items():
        start_coord = evader_symbol_coord_map[start_symbol]
        evader_goal_coords_map[start_coord] = [
            evader_symbol_coord_map[symbol] for symbol in goal_symbols
        ]

    return PEGrid(
        grid_width=width,
        grid_height=height,
        block_coords=block_coords,
        goal_coords_map=evader_goal_coords_map,
        evader_start_coords=evader_start_coords,
        pursuer_start_coords=pursuer_start_coords,
    )


# grid_name: (grid_make_fn, step_limit)
SUPPORTED_GRIDS = {
    "5x5": (get_5x5_grid, 20),
    "8x8": (get_8x8_grid, 50),
    "16x16": (get_16x16_grid, 100),
    "32x32": (get_32x32_grid, 200),
}


def load_grid(grid_name: str) -> PEGrid:
    """Load grid with given name."""
    grid_name = grid_name
    assert grid_name in SUPPORTED_GRIDS, (
        f"Unsupported grid name '{grid_name}'. Grid name must be one of: "
        f"{SUPPORTED_GRIDS.keys()}."
    )
    return SUPPORTED_GRIDS[grid_name][0]()



"""Module of wrapper classes.

Ref:
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/wrappers/__init__.py
"""
from posggym.wrappers.env_checker import PassiveEnvChecker
# from posggym.wrappers.flatten_observation import FlattenObservation
from posggym.wrappers.order_enforcing import OrderEnforcing
# from posggym.wrappers.record_env import RecordVideo
from posggym.wrappers.time_limit import TimeLimit



from gym import spaces

from posggym import ObservationWrapper, Env


class FlattenObservation(ObservationWrapper):
    """Observation wrapper that flattens the observation."""

    def __init__(self, env: Env):
        super().__init__(env)
        self.observation_spaces = tuple(
            spaces.flatten_space(observation_space)
            for observation_space in env.observation_spaces
        )

    def observations(self, observations):
        return tuple(
            spaces.flatten(self.env.observation_spaces[i], observations[i])
            for i in range(self.env.n_agents)
        )



"""The Two-Paths Grid World Environment.

An adversarial 2D grid world problem involving two agents, a runner and
a chaser. The runner's goal is to reach one of two goal location, with each
goal located at the end of a seperate path. The lengths of the two paths
have different lengths. The goal of the chaser is to intercept the runner
before it reaches a goal. The runner is considered caught if it is observed
by the chaser, or occupies the same location. However, the chaser is only
able to effectively cover one of the two goal locations.

This environment requires each agent to reason about the which path the
other agent will choose. It offers an ideal testbed for planning under
finite-nested reasoning assumptions since it is possible to map reasoning
level to the expected path choice.

Reference
---------
Schwartz, Jonathon, Ruijia Zhou, and Hanna Kurniawati. "Online Planning for
Interactive-POMDPs using Nested Monte Carlo Tree Search." In 2022 IEEE/RSJ International
Conference on Intelligent Robots and Systems (IROS), pp. 8770-8777. IEEE, 2022.

"""
import itertools
from typing import Dict, List, Optional, Set, SupportsFloat, Tuple, Union

from gymnasium import spaces

import posggym.envs.grid_world.render as render_lib
import posggym.model as M
from posggym.core import DefaultEnv
from posggym.envs.grid_world.core import Coord, Direction, Grid
from posggym.utils import seeding


TPState = Tuple[Coord, Coord]
TPAction = int
# Obs = (adj_obs, terminal)
TPObs = Tuple[Tuple[int, int, int, int], int]

# Cell obs
OPPONENT = 0
WALL = 1
EMPTY = 2

CELL_OBS = [OPPONENT, WALL, EMPTY]
CELL_OBS_STR = ["X", "#", "0"]


class TwoPathsEnv(DefaultEnv[TPState, TPObs, TPAction]):
    """The Two-Paths Grid World Environment.

    An adversarial 2D grid world problem involving two agents, a runner and
    a chaser. The runner's goal is to reach one of two goal location, with each
    goal located at the end of a seperate path. The lengths of the two paths
    have different lengths. The goal of the chaser is to intercept the runner
    before it reaches a goal. The runner is considered caught if it is observed
    by the chaser, or occupies the same location. However, the chaser is only
    able to effectively cover one of the two goal locations.

    This environment requires each agent to reason about the which path the
    other agent will choose. It offers an ideal testbed for planning under
    finite-nested reasoning assumptions since it is possible to map reasoning
    level to the expected path choice.

    The two agents start at opposite ends of the maps.

    Agents
    ------
    Runner=0
    Chaser=1

    State
    -----
    Each state contains the (x, y) (x=column, y=row, with origin at the
    top-left square of the grid) of the runner and chaser agent. Specifically,
    a states is ((x_runner, y_runner), (x_chaser, y_chaser))

    Actions
    -------
    Each agent has 4 actions corresponding to moving in the 4 cardinal
    directions (NORTH=0, EAST=1, SOUTH=2, WEST=3).

    Observation
    -----------
    Each agent observes the adjacent cells in the four cardinal directions and
    whether they are one of three things: OPPONENT=0, WALL=1, EMPTY=2.
    Each agent also observes whether a terminal state was reach (0/1). This is
    necessary for the infinite horizon model of the environment.

    Each observation is represented as a tuple:
        ((cell_north, cell_south, cell_east, cell_west), terminal)

    Reward
    ------
    Both agents receive a penalty of -0.01 for each step.
    If the runner reaches the goal then the runner recieves a reward of 1.0,
    while the chaser recieves a penalty of -1.0.
    If the runner is observed by the chaser, then the runner recieves a penalty
    of -1.0, while the chaser recieves a reward of 1.0.

    The rewards make the environment adversarial, but not strictly zero-sum,
    due to the small penalty each step.

    Transition Dynamics
    -------------------
    By default actions are deterministic and an episode ends when either the
    runner is caught, the runner reaches a goal, or the step limit is reached.

    The environment can also be run in stochastic mode by changing the
    action_probs parameter at initialization. This controls the probability
    the agent will move in the desired direction each step, otherwise moving
    randomly in one of the other 3 possible directions.

    Lastly, if using infinite_horizon mode then the environment resets to the
    start state once a terminal state is reached.
    """

    metadata = {
        "render_modes": ["human", "ansi", "rgb_array", "rgb_array_dict"],
        "render_fps": 4,
    }

    def __init__(
        self,
        grid_name: str,
        action_probs: Union[float, Tuple[float, float]] = 1.0,
        infinite_horizon: bool = False,
        render_mode: Optional[str] = None,
        **kwargs,
    ):
        self._viewer = None
        self._renderer: Optional[render_lib.GWRenderer] = None
        super().__init__(
            TwoPathsModel(grid_name, action_probs, infinite_horizon, **kwargs),
            render_mode=render_mode,
        )

    def render(self):
        grid = self.model.grid  # type: ignore
        if self.render_mode == "ansi":
            grid_str = grid.get_ascii_repr(self._state[0], self._state[1])

            output = [
                f"Step: {self._step_num}",
                grid_str,
            ]
            if self._last_actions is not None:
                action_str = ", ".join(
                    [str(Direction(a)) for a in self._last_actions.values()]
                )
                output.insert(1, f"Actions: <{action_str}>")
                output.append(f"Rewards: <{self._last_rewards}>")

            return "\n".join(output) + "\n"

        if self.render_mode == "human" and self._viewer is None:
            # pylint: disable=[import-outside-toplevel]
            from posggym.envs.grid_world import viewer

            self._viewer = viewer.GWViewer(
                "Two-Paths Env",
                (min(grid.width, 9), min(grid.height, 9)),
                num_agent_displays=len(self.possible_agents),
            )
            self._viewer.show(block=False)

        if self._renderer is None:
            static_objs = [
                render_lib.GWObject(coord, "green", render_lib.Shape.RECTANGLE)
                for coord in grid.goal_coords
            ]
            self._renderer = render_lib.GWRenderer(
                len(self.possible_agents), grid, static_objs, render_blocks=True
            )

        agent_coords = self._state
        agent_obs_coords = tuple(
            grid.get_neighbours(self._state[i], True)
            for i in range(len(self.possible_agents))
        )
        for i, coord in enumerate(agent_coords):
            agent_obs_coords[i].append(coord)
        agent_dirs = tuple(Direction.NORTH for _ in range(len(self.possible_agents)))

        env_img = self._renderer.render(
            self._state,
            agent_obs_coords,
            agent_dirs,
            other_objs=None,
            agent_colors=None,
        )
        agent_obs_imgs = self._renderer.render_all_agent_obs(
            env_img,
            agent_coords,
            agent_dirs,
            agent_obs_dims=(1, 1, 1),
            out_of_bounds_obj=render_lib.GWObject(
                (0, 0), "grey", render_lib.Shape.RECTANGLE
            ),
            agent_obs_coords=agent_obs_coords,
        )

        if self.render_mode == "human":
            self._viewer.update_img(env_img, agent_idx=None)
            for i, obs_img in enumerate(agent_obs_imgs):
                self._viewer.update_img(obs_img, agent_idx=i)
            self._viewer.display_img()
        elif self.render_mode == "rgb_array":
            return env_img
        else:
            # rgb_array_dict
            return dict(enumerate(agent_obs_imgs))

    def close(self) -> None:
        if self._viewer is not None:
            self._viewer.close()
            self._viewer = None


class TwoPathsModel(M.POSGModel[TPState, TPObs, TPAction]):
    """Two-Paths Problem Model.

    Parameters
    ----------
    grid_name : str
        determines the environment map (see posggym.envs.two_paths.grid for
        available grids)
    action_probs : float or (float, float)
        the action success probability for each agent. By default the
        environment is deterministic (action_probs=1.0).
    infinite_horizon : bool
        whether problem should terminate once a terminal state is reached
        (default, False) or reset to start position and continue (True).

    """

    NUM_AGENTS = 2

    RUNNER_IDX = 0
    CHASER_IDX = 1

    R_ACTION = -0.01
    R_CAPTURE = -1.0  # Runner reward, Chaser = -R_CAPTURE
    R_SAFE = 1.0  # Runner reward, Chaser = -R_SAFE

    def __init__(
        self,
        grid_name: str,
        action_probs: Union[float, Tuple[float, float]] = 1.0,
        infinite_horizon: bool = False,
        **kwargs,
    ):
        self.grid = load_grid(grid_name)
        self._infinite_horizon = infinite_horizon
        if isinstance(action_probs, float):
            action_probs = (action_probs, action_probs)
        self._action_probs = action_probs

        self.possible_agents = tuple(range(self.NUM_AGENTS))
        self.state_space = spaces.Tuple(
            tuple(
                spaces.Tuple(
                    (
                        spaces.Discrete(self.grid.width),
                        spaces.Discrete(self.grid.height),
                    )
                )
                for _ in range(len(self.possible_agents))
            )
        )
        self.action_spaces = {
            i: spaces.Discrete(len(Direction)) for i in self.possible_agents
        }
        self.observation_spaces = {
            i: spaces.Tuple(
                (
                    spaces.Tuple(
                        (
                            spaces.Discrete(len(CELL_OBS)),
                            spaces.Discrete(len(CELL_OBS)),
                            spaces.Discrete(len(CELL_OBS)),
                            spaces.Discrete(len(CELL_OBS)),
                        )
                    ),
                    spaces.Discrete(2),
                )
            )
            for i in self.possible_agents
        }
        self.observation_first = True
        self.is_symmetric = False

    @property
    def reward_ranges(self) -> Dict[M.AgentID, Tuple[SupportsFloat, SupportsFloat]]:
        return {i: (self.R_CAPTURE, self.R_SAFE) for i in self.possible_agents}

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: TPState) -> List[M.AgentID]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> TPState:
        return (self.grid.init_runner_coord, self.grid.init_chaser_coord)

    def sample_agent_initial_state(self, agent_id: M.AgentID, obs: TPObs) -> TPState:
        return self.sample_initial_state()

    def get_initial_belief_dist(self) -> Dict[TPState, float]:
        s_0 = (self.grid.init_runner_coord, self.grid.init_chaser_coord)
        return {
            s: float(s == s_0)  # type: ignore
            for s in itertools.product(self.grid.all_coords, repeat=2)
        }

    def sample_initial_obs(self, state: TPState) -> Dict[M.AgentID, TPObs]:
        return self._get_obs(state, False)

    def step(
        self, state: TPState, actions: Dict[M.AgentID, TPAction]
    ) -> M.JointTimestep[TPState, TPObs]:
        assert all(0 <= a_i < len(Direction) for a_i in actions.values())
        next_state = self._get_next_state(state, actions)
        rewards = self._get_rewards(next_state)
        terminal = self._state_is_terminal(next_state)

        info: Dict[M.AgentID, Dict] = {i: {} for i in self.possible_agents}
        if terminal:
            for i, outcome in self._get_outcome(next_state).items():
                info[i]["outcome"] = outcome

        if self._infinite_horizon and terminal:
            next_state = self.sample_initial_state()
            terminal = False

        obs = self._get_obs(next_state, terminal)
        truncated = {i: False for i in self.possible_agents}
        terminated = {i: terminal for i in self.possible_agents}

        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, terminal, info
        )

    def _get_next_state(
        self, state: TPState, actions: Dict[M.AgentID, TPAction]
    ) -> TPState:
        runner_coord = state[self.RUNNER_IDX]
        chaser_coord = state[self.CHASER_IDX]
        runner_a = actions[self.RUNNER_IDX]
        chaser_a = actions[self.CHASER_IDX]

        if self.rng.random() > self._action_probs[self.CHASER_IDX]:
            other_as = [a for a in Direction if a != chaser_a]
            chaser_a = self.rng.choice(other_as)
        chaser_next_coord = self.grid.get_next_coord(
            chaser_coord, Direction(chaser_a), ignore_blocks=False
        )

        if chaser_next_coord == runner_coord:
            # Runner considered to capture Fugitive
            runner_next_coord = runner_coord
        else:
            if self.rng.random() > self._action_probs[self.RUNNER_IDX]:
                other_as = [a for a in Direction if a != runner_a]
                runner_a = self.rng.choice(other_as)
            runner_next_coord = self.grid.get_next_coord(
                runner_coord, Direction(runner_a), ignore_blocks=False
            )

        return (runner_next_coord, chaser_next_coord)

    def _get_obs(self, state: TPState, terminal: bool) -> Dict[M.AgentID, TPObs]:
        runner_coord = state[self.RUNNER_IDX]
        chaser_coord = state[self.CHASER_IDX]
        return {
            self.RUNNER_IDX: (
                self._get_adj_obs(runner_coord, chaser_coord),
                int(terminal),
            ),
            self.CHASER_IDX: (
                self._get_adj_obs(chaser_coord, runner_coord),
                int(terminal),
            ),
        }

    def _get_adj_obs(
        self, coord: Coord, opponent_coord: Coord
    ) -> Tuple[int, int, int, int]:
        adj_obs = []
        for d in Direction:
            next_coord = self.grid.get_next_coord(coord, d, False)
            if next_coord == opponent_coord:
                adj_obs.append(OPPONENT)
            elif next_coord == coord:
                adj_obs.append(WALL)
            else:
                adj_obs.append(EMPTY)
        return tuple(adj_obs)  # type: ignore

    def _get_rewards(self, state: TPState) -> Dict[M.AgentID, SupportsFloat]:
        runner_coord = state[self.RUNNER_IDX]
        chaser_coord = state[self.CHASER_IDX]
        r_runner, r_chaser = (self.R_ACTION, self.R_ACTION)
        if runner_coord in self.grid.goal_coords:
            r_runner, r_chaser = (self.R_SAFE, -self.R_SAFE)
        if runner_coord == chaser_coord or runner_coord in self.grid.get_neighbours(
            chaser_coord, True
        ):
            r_runner, r_chaser = (self.R_CAPTURE, -self.R_CAPTURE)
        return {self.RUNNER_IDX: r_runner, self.CHASER_IDX: r_chaser}

    def _get_outcome(self, state: TPState) -> Dict[M.AgentID, M.Outcome]:
        if self._infinite_horizon:
            return {self.RUNNER_IDX: M.Outcome.NA, self.CHASER_IDX: M.Outcome.NA}

        # Assuming state is terminal
        runner_coord = state[self.RUNNER_IDX]
        chaser_coord = state[self.CHASER_IDX]
        runner_outcome, chaser_outcome = (M.Outcome.DRAW, M.Outcome.DRAW)
        if runner_coord in self.grid.goal_coords:
            runner_outcome, chaser_outcome = (M.Outcome.WIN, M.Outcome.LOSS)
        elif runner_coord == chaser_coord or runner_coord in self.grid.get_neighbours(
            chaser_coord, True
        ):
            runner_outcome, chaser_outcome = (M.Outcome.LOSS, M.Outcome.WIN)
        return {self.RUNNER_IDX: runner_outcome, self.CHASER_IDX: chaser_outcome}

    def _state_is_terminal(self, state: TPState) -> bool:
        runner_coords = state[self.RUNNER_IDX]
        chaser_coords = state[self.CHASER_IDX]
        if chaser_coords == runner_coords or runner_coords in self.grid.goal_coords:
            return True
        neighbour_coords = self.grid.get_neighbours(chaser_coords, True)
        return runner_coords in neighbour_coords


class TPGrid(Grid):
    """A grid for the Two-Paths Problem."""

    def __init__(
        self,
        grid_width: int,
        grid_height: int,
        block_coords: Set[Coord],
        goal_coords: Set[Coord],
        init_runner_coord: Coord,
        init_chaser_coord: Coord,
    ):
        super().__init__(grid_width, grid_height, block_coords)
        self.goal_coords = goal_coords
        self.init_runner_coord = init_runner_coord
        self.init_chaser_coord = init_chaser_coord

    def get_ascii_repr(
        self, runner_coord: Optional[Coord], chaser_coord: Optional[Coord]
    ) -> str:
        """Get ascii repr of grid."""
        grid_repr = []
        for row in range(self.height):
            row_repr = []
            for col in range(self.width):
                coord = (col, row)
                if coord in self.goal_coords:
                    row_repr.append("G")
                elif coord in self.block_coords:
                    row_repr.append("#")
                else:
                    row_repr.append(".")
            grid_repr.append(row_repr)

        if runner_coord is not None:
            grid_repr[runner_coord[1]][runner_coord[0]] = "R"
        if chaser_coord is not None:
            if chaser_coord == runner_coord:
                grid_repr[chaser_coord[1]][chaser_coord[0]] = "X"
            else:
                grid_repr[chaser_coord[1]][chaser_coord[0]] = "C"

        return "\n".join(list(list((" ".join(r) for r in grid_repr))))

    def get_init_ascii_repr(self) -> str:
        """Get ascii repr of initial grid."""
        return self.get_ascii_repr(self.init_runner_coord, self.init_chaser_coord)


def get_3x3_grid() -> TPGrid:
    """Generate the Two-Paths 3-by-3 grid layout.

       012
      #####
    0 #GC #
    1 # #G#
    2 # R #
      #####

    """
    return TPGrid(
        grid_height=3,
        grid_width=3,
        block_coords=set([(1, 1)]),
        goal_coords=set([(0, 0), (2, 1)]),
        init_runner_coord=(1, 2),
        init_chaser_coord=(1, 0),
    )


def get_4x4_grid() -> TPGrid:
    """Generate the Choose a Path 4-by-4 grid layout.

       0123
      ######
    0 #G C #
    1 # ##G#
    2 # #  #
    3 #  R##
      ######

    """
    block_coords = set(
        [
            #
            (1, 1),
            (2, 1),
            (1, 2),
            (3, 3),
        ]
    )

    return TPGrid(
        grid_height=4,
        grid_width=4,
        block_coords=block_coords,
        goal_coords=set([(0, 0), (3, 1)]),
        init_runner_coord=(2, 3),
        init_chaser_coord=(2, 0),
    )


def get_7x7_grid() -> TPGrid:
    """Generate the Two-Paths 7-by-7 grid layout.

       0123456
      #########
    0 #G   C  #
    1 # ##### #
    2 # #### G#
    3 #  ### ##
    4 ## ### ##
    5 ##  #  ##
    6 ### R ###
      #########

    """
    block_coords = set(
        [
            #
            (1, 1),
            (2, 1),
            (3, 1),
            (4, 1),
            (5, 1),
            (1, 2),
            (2, 2),
            (3, 2),
            (4, 2),
            (2, 3),
            (3, 3),
            (4, 3),
            (6, 3),
            (0, 4),
            (2, 4),
            (3, 4),
            (4, 4),
            (6, 4),
            (0, 5),
            (3, 5),
            (6, 5),
            (0, 6),
            (1, 6),
            (5, 6),
            (6, 6),
        ]
    )

    return TPGrid(
        grid_height=7,
        grid_width=7,
        block_coords=block_coords,
        goal_coords=set([(0, 0), (6, 2)]),
        init_runner_coord=(3, 6),
        init_chaser_coord=(4, 0),
    )


# grid_name: (grid_make_fn, episode step_limit)
SUPPORTED_GRIDS = {
    "3x3": (get_3x3_grid, 20),
    "4x4": (get_4x4_grid, 20),
    "7x7": (get_7x7_grid, 20),
}


def load_grid(grid_name: str) -> TPGrid:
    """Load grid with given name."""
    grid_name = grid_name
    assert grid_name in SUPPORTED_GRIDS, (
        f"Unsupported grid name '{grid_name}'. Grid name must be one of: "
        f"{SUPPORTED_GRIDS.keys()}."
    )
    return SUPPORTED_GRIDS[grid_name][0]()



"""The Driving Grid World Environment.

A general-sum 2D grid world problem involving multiple agents. Each agent
controls a vehicle and is tasked with driving the vehicle from it's start
location to a destination location while avoiding crashing into obstacles
or other vehicles.

This environment requires each agent to navigate in the world while also
taking care to avoid crashing into other players. The dynamics and
observations of the environment are such that avoiding collisions requires
some planning in order for the vehicle to brake in time or maintain a good
speed. Depending on the grid layout, the environment will require agents to
reason about and possibly coordinate with the other vehicles.

References
----------
- Adam Lerer and Alexander Peysakhovich. 2019. Learning Existing Social Conventions via
Observationally Augmented Self-Play. In Proceedings of the 2019 AAAI/ACM Conference on
AI, Ethics, and Society. 107–114.
- Kevin R. McKee, Joel Z. Leibo, Charlie Beattie, and Richard Everett. 2022.
Quantifying the Effects of Environment and Population Diversity in Multi-Agent
Reinforcement Learning. Autonomous Agents and Multi-Agent Systems 36, 1 (2022), 1–16

"""
import enum
from itertools import product
from typing import (
    Any,
    Dict,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    SupportsFloat,
    Tuple,
    Union,
)

from gymnasium import spaces

import posggym.envs.grid_world.render as render_lib
import posggym.model as M
from posggym.core import DefaultEnv
from posggym.envs.grid_world.core import (
    DIRECTION_ASCII_REPR,
    Coord,
    Direction,
    Grid,
    GridCycler,
    GridGenerator,
)
from posggym.utils import seeding


class Speed(enum.IntEnum):
    """A speed setting for a vehicle."""

    REVERSE = 0
    STOPPED = 1
    FORWARD_SLOW = 2
    FORWARD_FAST = 3


class CollisionType(enum.IntEnum):
    """Type of collision for a vehicle."""

    NONE = 0
    OBSTACLE = 1
    VEHICLE = 2


class VehicleState(NamedTuple):
    """The state of a vehicle in the Driving Environment."""

    coord: Coord
    facing_dir: Direction
    speed: Speed
    dest_coord: Coord
    dest_reached: int
    crashed: int
    min_dest_dist: int


DState = Tuple[VehicleState, ...]

# Initial direction and speed of each vehicle
INIT_DIR = Direction.NORTH
INIT_SPEED = Speed.STOPPED

# The actions
DAction = int
DO_NOTHING = 0
ACCELERATE = 1
DECELERATE = 2
TURN_RIGHT = 3
TURN_LEFT = 4

ACTIONS = [DO_NOTHING, ACCELERATE, DECELERATE, TURN_RIGHT, TURN_LEFT]
ACTIONS_STR = ["0", "acc", "dec", "tr", "tl"]

# Obs = (adj_obs, speed, dest_coord, dest_reached, crashed)
DObs = Tuple[Tuple[int, ...], Speed, Coord, int, int]

# Cell obs
VEHICLE = 0
WALL = 1
EMPTY = 2
DESTINATION = 3

CELL_OBS = [VEHICLE, WALL, EMPTY, DESTINATION]
CELL_OBS_STR = ["V", "#", "0", "D"]


class DrivingEnv(DefaultEnv[DState, DObs, DAction]):
    """The Driving Grid World Environment.

    A general-sum 2D grid world problem involving multiple agents. Each agent
    controls a vehicle and is tasked with driving the vehicle from it's start
    location to a destination location while avoiding crashing into obstacles
    or other vehicles.

    This environment requires each agent to navigate in the world while also
    taking care to avoid crashing into other players. The dynamics and
    observations of the environment are such that avoiding collisions requires
    some planning in order for the vehicle to brake in time or maintain a good
    speed. Depending on the grid layout, the environment will require agents to
    reason about and possibly coordinate with the other vehicles.


    Agents
    ------
    Varied number

    State
    -----
    Each state is made up of the state of each vehicle, which in turn is
    defined by:

    - the (x, y) (x=column, y=row, with origin at the top-left square of the
      grid) of the vehicle,
    - the direction the vehicle is facing NORTH=0, SOUTH=1, EAST=2, WEST=3),
    - the speed of the vehicle (REVERSE=-1, STOPPED=0, FORWARD_SLOW=1,
      FORWARD_FAST=2),
    - the (x, y) of the vehicles destination
    - whether the vehicle has reached it's destination or not
    - whether the vehicle has crashed or not

    Actions
    -------
    Each agent has 5 actions: DO_NOTHING=0, ACCELERATE=1, DECELERATE=2,
    TURN_RIGHT=3, TURN_LEFT=4

    Observation
    -----------
    Each agent observes their current speed along with the cells in their local
    area. The size of the local area observed is controlled by the `obs_dims`
    parameter. For each cell in the observed are the agent observes whether
    they are one of four things: VEHICLE=0, WALL=1, EMPTY=2, DESTINATION=3.
    Each agent also observes the (x, y) coordinates of their destination,
    whether they have reached the destination, and whether they have crashed.

    Each observation is represented as a tuple:
        ((local obs), speed, destination coord, destination reached, crashed)

    Reward
    ------
    All agents receive a penalty of 0.00 for each step. They also recieve a
    penalty of -0.5 for hitting an obstacle (if ``obstacle_collision=True``),
    and -1.0 for hitting another vehicle. A reward of 1.0 is given if the agent
    reaches it's destination.

    Transition Dynamics
    -------------------
    Actions are deterministic and movement is determined by direction the
    vehicle is facing and it's speed:

    - Speed=-1 (REVERSE) - vehicle moves one cell in the opposite direction
    - Speed=0 (STOPPED) - vehicle remains in same cell
    - Speed=1 (FORWARD_SLOW) - vehicle move one cell in facing direction
    - Speed=1 (FORWARD_FAST) - vehicle moves two cells in facing direction

    Accelerating increases speed by 1, while deceleration decreased speed by 1.
    If the vehicle will hit a wall or an other vehicle when moving from one
    cell to another then it remains in it's current cell and its crashed state
    variable is updated. Once a vehicle reaches it's destination it is stuck.

    Episodes ends when all agents have either reached their destination or
    crashed, or the episode step limit is reached.

    References
    ----------
    - Adam Lerer and Alexander Peysakhovich. 2019. Learning Existing Social Conventions
    via Observationally Augmented Self-Play. In Proceedings of the 2019 AAAI/ACM
    Conference on AI, Ethics, and Society. 107–114.
    - Kevin R. McKee, Joel Z. Leibo, Charlie Beattie, and Richard Everett. 2022.
    Quantifying the Effects of Environment and Population Diversity in Multi-Agent
    Reinforcement Learning. Autonomous Agents and Multi-Agent Systems 36, 1 (2022), 1–16

    """

    metadata = {
        "render_modes": ["human", "ansi", "rgb_array", "rgb_array_dict"],
        "render_fps": 4,
    }

    def __init__(
        self,
        grid: Union[str, "DrivingGrid"],
        num_agents: int,
        obs_dim: Tuple[int, int, int],
        obstacle_collisions: bool,
        render_mode: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            DrivingModel(grid, num_agents, obs_dim, obstacle_collisions, **kwargs),
            render_mode=render_mode,
        )
        self._obs_dim = obs_dim
        self._viewer = None
        self._renderer: Optional[render_lib.GWRenderer] = None
        # Whether to re-render blocks each new episode
        # This is needed when block positions change between episodes
        self._rerender_blocks = False

    def render(self):
        model: DrivingModel = self.model  # type: ignore
        if self.render_mode == "ansi":
            grid_str = model.grid.get_ascii_repr(
                [vs.coord for vs in self._state],
                [vs.facing_dir for vs in self._state],
                [vs.dest_coord for vs in self._state],
            )

            output = [
                f"Step: {self._step_num}",
                grid_str,
            ]
            if self._last_actions is not None:
                action_str = ", ".join(
                    [ACTIONS_STR[a] for a in self._last_actions.values()]
                )
                output.insert(1, f"Actions: <{action_str}>")
                output.append(f"Rewards: <{self._last_rewards}>")

            return "\n".join(output) + "\n"

        if self.render_mode == "human" and self._viewer is None:
            # pylint: disable=[import-outside-toplevel]
            from posggym.envs.grid_world import viewer

            self._viewer = viewer.GWViewer(  # type: ignore
                "Driving Env",
                (min(model.grid.width, 9), min(model.grid.height, 9)),
                num_agent_displays=len(self.possible_agents),
            )
            self._viewer.show(block=False)  # type: ignore

        blocks_added = False
        if self._renderer is None:
            blocks_added = True
            self._renderer = render_lib.GWRenderer(
                len(self.possible_agents), model.grid, [], render_blocks=True
            )

        agent_obs_coords = tuple(
            model.get_obs_coords(vs.coord, vs.facing_dir) for vs in self._state
        )
        agent_coords = tuple(vs.coord for vs in self._state)
        agent_dirs = tuple(vs.facing_dir for vs in self._state)

        # Add agent destination locations
        other_objs = [
            render_lib.GWObject(
                vs.dest_coord,
                render_lib.get_agent_color(i),
                render_lib.Shape.RECTANGLE,
                # make dest squares slightly different to vehicle color
                alpha=0.2,
            )
            for i, vs in enumerate(self._state)
        ]
        # Add blocks, as necessary
        if self._rerender_blocks and not blocks_added:
            for block_coord in model.grid.block_coords:
                other_objs.append(
                    render_lib.GWObject(block_coord, "grey", render_lib.Shape.RECTANGLE)
                )
        # Add visualization for crashed agents
        for i, vs in enumerate(self._state):
            if vs.crashed:
                other_objs.append(
                    render_lib.GWObject(vs.coord, "yellow", render_lib.Shape.CIRCLE)
                )

        env_img = self._renderer.render(
            agent_coords,
            agent_obs_coords,
            agent_dirs=agent_dirs,
            other_objs=other_objs,
            agent_colors=None,
        )
        agent_obs_imgs = self._renderer.render_all_agent_obs(
            env_img,
            agent_coords,
            agent_dirs,
            agent_obs_dims=self._obs_dim,
            out_of_bounds_obj=render_lib.GWObject(
                (0, 0), "grey", render_lib.Shape.RECTANGLE
            ),
        )

        if self.render_mode == "human":
            self._viewer.update_img(env_img, agent_idx=None)
            for i, obs_img in enumerate(agent_obs_imgs):
                self._viewer.update_img(obs_img, agent_idx=i)
            self._viewer.display_img()
        elif self.render_mode == "rgb_array":
            return env_img
        else:
            # rgb_array_dict
            return dict(enumerate(agent_obs_imgs))

    def close(self) -> None:
        if self._viewer is not None:
            self._viewer.close()
            self._viewer = None


class DrivingGenEnv(DrivingEnv):
    """The Generated Driving Grid World Environment.

    This is the same as the Driving Environment except that a new grid is
    generated at each reset.

    The generated Grids can be:

    1. set to be from some list of grids, in which case the grids in the list
       will be cycled through (in the same order or shuffled)
    2. generated a new each reset. Depending on grid size and generator
       parameters this will lead to possibly unique grids each episode.

    The seed parameter can be used to ensure the same grids are used on
    different runs.

    """

    def __init__(
        self,
        num_agents: int,
        obs_dim: Tuple[int, int, int],
        obstacle_collisions: bool,
        n_grids: Optional[int],
        generator_params: Dict[str, Any],
        shuffle_grid_order: bool = True,
        seed: Optional[int] = None,
        **kwargs,
    ):
        if "seed" not in generator_params:
            generator_params["seed"] = seed

        self._n_grids = n_grids
        self._gen_params = generator_params
        self._shuffle_grid_order = shuffle_grid_order
        self._gen = DrivingGridGenerator(**generator_params)
        # Need to re-add blocks each time since these change each episode
        self._rerender_blocks = True

        if n_grids is not None:
            grids = self._gen.generate_n(n_grids)
            self._cycler = GridCycler(grids, shuffle_grid_order, seed=seed)
            grid: "DrivingGrid" = grids[0]  # type: ignore
        else:
            self._cycler = None  # type: ignore
            grid = self._gen.generate()

        self._model_kwargs = {
            "num_agents": num_agents,
            "obs_dim": obs_dim,
            "obstacle_collisions": obstacle_collisions,
            **kwargs,
        }

        super().__init__(
            grid,
            num_agents,
            obs_dim,
            obstacle_collisions=obstacle_collisions,
            seed=seed,
            **kwargs,
        )

    def reset(
        self, *, seed: Optional[int] = None, options: Optional[Dict[str, Any]] = None
    ) -> Tuple[Optional[Dict[M.AgentID, DObs]], Dict[M.AgentID, Dict]]:
        if seed is not None:
            self._gen_params["seed"] = seed
            self._gen = DrivingGridGenerator(**self._gen_params)

            if self._n_grids is not None:
                grids = self._gen.generate_n(self._n_grids)
                self._cycler = GridCycler(grids, self._shuffle_grid_order, seed=seed)

        if self._n_grids:
            grid = self._cycler.next()
        else:
            grid = self._gen.generate()

        self.model = DrivingModel(
            grid,  # type: ignore
            **self._model_kwargs,  # type: ignore
        )

        return super().reset(seed=seed)


class DrivingModel(M.POSGModel[DState, DObs, DAction]):
    """Driving Problem Model.

    Parameters
    ----------
    grid : DrivingGrid
        the grid environment for the model scenario
    num_agents : int
        the number of agents in the model scenario
    obs_dims : (int, int, int)
        number of cells in front, behind, and to the side that each agent
        can observe
    obstacle_collisions : bool
        whether cars can crash into wall and other obstacles, on top of
        crashing into other vehicles

    """

    R_STEP_COST = 0.00
    R_CRASH_OBJECT = -0.05
    R_CRASH_VEHICLE = -1.0
    R_DESTINATION_REACHED = 1.0
    R_PROGRESS = 0.05

    def __init__(
        self,
        grid: Union[str, "DrivingGrid"],
        num_agents: int,
        obs_dim: Tuple[int, int, int],
        obstacle_collisions: bool,
        **kwargs,
    ):
        if isinstance(grid, str):
            grid = load_grid(grid)

        assert 0 < num_agents <= grid.supported_num_agents
        assert obs_dim[0] > 0 and obs_dim[1] >= 0 and obs_dim[2] >= 0
        self.grid = grid
        self._obs_front, self._obs_back, self._obs_side = obs_dim
        self._obstacle_collisions = obstacle_collisions

        def _coord_space():
            return spaces.Tuple(
                (spaces.Discrete(self.grid.width), spaces.Discrete(self.grid.height))
            )

        self.possible_agents = tuple(range(num_agents))
        self.state_space = spaces.Tuple(
            tuple(
                spaces.Tuple(
                    (
                        _coord_space(),
                        spaces.Discrete(len(Direction)),
                        spaces.Discrete(len(Speed)),
                        _coord_space(),  # destination coord
                        spaces.Discrete(2),  # destination reached
                        spaces.Discrete(2),  # crashed
                        # min distance to destination
                        # set this to upper bound of min shortest path distance, so
                        # state space works for generated grids as well
                        spaces.Discrete(self.grid.width * self.grid.height),
                    )
                )
                for _ in range(len(self.possible_agents))
            )
        )
        self.action_spaces = {
            i: spaces.Discrete(len(ACTIONS)) for i in self.possible_agents
        }
        obs_depth = self._obs_front + self._obs_back + 1
        obs_width = (2 * self._obs_side) + 1
        self.observation_spaces = {
            i: spaces.Tuple(
                (
                    spaces.Tuple(
                        tuple(
                            spaces.Discrete(len(CELL_OBS))
                            for _ in range(obs_depth * obs_width)
                        )
                    ),
                    spaces.Discrete(len(Speed)),
                    _coord_space(),  # dest coord,
                    spaces.Discrete(2),  # dest reached
                    spaces.Discrete(2),  # crashed
                )
            )
            for i in self.possible_agents
        }
        self.observation_first = True
        self.is_symmetric = True

    @property
    def reward_ranges(self) -> Dict[M.AgentID, Tuple[SupportsFloat, SupportsFloat]]:
        return {
            i: (self.R_CRASH_VEHICLE, self.R_DESTINATION_REACHED)
            for i in self.possible_agents
        }

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: DState) -> List[M.AgentID]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> DState:
        state = []
        chosen_start_coords: Set[Coord] = set()
        chosen_dest_coords: Set[Coord] = set()
        for i in range(len(self.possible_agents)):
            start_coords_i = self.grid.start_coords[i]
            avail_start_coords = start_coords_i.difference(chosen_start_coords)
            start_coord = self.rng.choice(list(avail_start_coords))
            chosen_start_coords.add(start_coord)

            dest_coords_i = self.grid.dest_coords[i]
            avail_dest_coords = dest_coords_i.difference(chosen_dest_coords)
            if start_coord in avail_dest_coords:
                avail_dest_coords.remove(start_coord)
            dest_coord = self.rng.choice(list(avail_dest_coords))
            chosen_dest_coords.add(dest_coord)

            dest_dist = self.grid.get_shortest_path_distance(start_coord, dest_coord)

            state_i = VehicleState(
                coord=start_coord,
                facing_dir=INIT_DIR,
                speed=INIT_SPEED,
                dest_coord=dest_coord,
                dest_reached=int(False),
                crashed=int(False),
                min_dest_dist=dest_dist,
            )
            state.append(state_i)
        return tuple(state)

    def sample_agent_initial_state(self, agent_id: M.AgentID, obs: DObs) -> DState:
        agent_id = int(agent_id)
        possible_agent_start_coords = set()
        agent_dest_coords = obs[2]

        # Need to get start states for agent that are valid given initial obs
        # Need to handle possible start states for other agents
        for all_agent_start_coords in product(
            *[list(s) for s in self.grid.start_coords[: len(self.possible_agents)]]
        ):
            if len(set(all_agent_start_coords)) != len(all_agent_start_coords):
                # skip any sets of start coord that contain duplicates
                continue
            local_obs = self._get_local_cell__obs(
                agent_id, all_agent_start_coords, INIT_DIR, agent_dest_coords
            )
            if local_obs == obs[0]:
                possible_agent_start_coords.add(all_agent_start_coords[agent_id])

        state = []
        chosen_start_coords: Set[Coord] = set()
        chosen_dest_coords: Set[Coord] = set()

        agent_start_coord = self.rng.choice(list(possible_agent_start_coords))
        chosen_start_coords.add(agent_start_coord)
        chosen_dest_coords.add(agent_dest_coords)

        for i in range(len(self.possible_agents)):
            if i == agent_id:
                start_coord = agent_start_coord
            else:
                start_coords_i = self.grid.start_coords[i]
                avail_coords = start_coords_i.difference(chosen_start_coords)
                start_coord = self.rng.choice(list(avail_coords))
                chosen_start_coords.add(start_coord)

            if i == agent_id:
                dest_coord = agent_dest_coords
            else:
                dest_coords_i = self.grid.dest_coords[i]
                avail_coords = dest_coords_i.difference(chosen_dest_coords)
                if start_coord in avail_coords:
                    avail_coords.remove(start_coord)
                dest_coord = self.rng.choice(list(avail_coords))
                chosen_dest_coords.add(dest_coord)

            dest_dist = self.grid.get_shortest_path_distance(start_coord, dest_coord)

            state_i = VehicleState(
                coord=start_coord,
                facing_dir=INIT_DIR,
                speed=INIT_SPEED,
                dest_coord=dest_coord,
                dest_reached=int(False),
                crashed=int(False),
                min_dest_dist=dest_dist,
            )
            state.append(state_i)
        return tuple(state)

    def sample_initial_obs(self, state: DState) -> Dict[M.AgentID, DObs]:
        return self._get_obs(state)

    def step(
        self, state: DState, actions: Dict[M.AgentID, DAction]
    ) -> M.JointTimestep[DState, DObs]:
        assert all(a_i in ACTIONS for a_i in actions.values())
        next_state, collision_types = self._get_next_state(state, actions)
        obs = self._get_obs(next_state)
        rewards = self._get_rewards(state, next_state, collision_types)
        terminated = {
            i: bool(next_state[int(i)].dest_reached or next_state[int(i)].crashed)
            for i in self.possible_agents
        }
        truncated = {i: False for i in self.possible_agents}
        all_done = all(terminated.values())

        info: Dict[M.AgentID, Dict] = {i: {} for i in self.possible_agents}
        for i in range(len(self.possible_agents)):
            if next_state[i].dest_reached:
                outcome_i = M.Outcome.WIN
            elif next_state[i].crashed:
                outcome_i = M.Outcome.LOSS
            else:
                outcome_i = M.Outcome.NA
            info[i]["outcome"] = outcome_i

        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _get_next_state(
        self, state: DState, actions: Dict[M.AgentID, DAction]
    ) -> Tuple[DState, List[CollisionType]]:
        exec_order = list(range(len(self.possible_agents)))
        self.rng.shuffle(exec_order)

        next_state = [vs for vs in state]
        vehicle_coords: Set[Coord] = set([vs.coord for vs in state])
        collision_types = [CollisionType.NONE] * len(self.possible_agents)

        for i in exec_order:
            action_i = actions[i]
            state_i = state[i]
            next_state_i = next_state[i]
            if state_i.dest_reached or next_state_i.crashed:
                # already at destination or crashed, or was crashed into this
                # step
                continue

            vehicle_coords.remove(state_i.coord)

            next_speed = self._get_next_speed(action_i, state_i.speed)
            move_dir = self._get_move_direction(
                action_i, next_speed, state_i.facing_dir
            )
            next_dir = self._get_next_direction(action_i, state_i.facing_dir)
            next_speed = self._get_next_speed(action_i, state_i.speed)
            next_coord, collision_type, hit_vehicle = self._get_next_coord(
                state_i.coord, next_speed, move_dir, vehicle_coords
            )

            min_dest_dist = min(
                state_i.min_dest_dist,
                self.grid.get_shortest_path_distance(next_coord, state_i.dest_coord),
            )

            crashed = False
            if collision_type == CollisionType.VEHICLE:
                # update state of vehicle that was hit
                crashed = True
                collision_types[i] = collision_type
                for j in range(len(self.possible_agents)):
                    next_state_j = next_state[j]
                    if next_state_j.coord == hit_vehicle:
                        collision_types[j] = CollisionType.VEHICLE
                        next_state[j] = VehicleState(
                            coord=next_state_j.coord,
                            facing_dir=next_state_j.facing_dir,
                            speed=next_state_j.speed,
                            dest_coord=next_state_j.dest_coord,
                            dest_reached=next_state_j.dest_reached,
                            crashed=int(True),
                            min_dest_dist=next_state_j.min_dest_dist,
                        )
                        break
            elif collision_type == CollisionType.OBSTACLE:
                crashed = self._obstacle_collisions
                collision_types[i] = collision_type

            next_state[i] = VehicleState(
                coord=next_coord,
                facing_dir=next_dir,
                speed=next_speed,
                dest_coord=state_i.dest_coord,
                dest_reached=int(next_coord == state_i.dest_coord),
                crashed=int(crashed),
                min_dest_dist=min_dest_dist,
            )

            vehicle_coords.add(next_coord)

        return tuple(next_state), collision_types

    def _reset_vehicle(
        self, v_idx: int, vs_i: VehicleState, vehicle_coords: Set[Coord]
    ) -> VehicleState:
        if not (vs_i.dest_reached or vs_i.crashed):
            return vs_i

        start_coords_i = self.grid.start_coords[v_idx]
        avail_start_coords = start_coords_i.difference(vehicle_coords)

        if vs_i.coord in start_coords_i:
            # add it back in since it will be remove during difference op
            avail_start_coords.add(vs_i.coord)

        new_coord = self.rng.choice(list(avail_start_coords))

        min_dest_dist = self.grid.manhattan_dist(new_coord, vs_i.dest_coord)

        new_vs_i = VehicleState(
            coord=new_coord,
            facing_dir=INIT_DIR,
            speed=INIT_SPEED,
            dest_coord=vs_i.dest_coord,
            dest_reached=int(False),
            crashed=int(False),
            min_dest_dist=min_dest_dist,
        )
        return new_vs_i

    def _get_move_direction(
        self, action: DAction, speed: Speed, curr_dir: Direction
    ) -> Direction:
        if speed == Speed.REVERSE:
            # No turning while in reverse,
            # so movement dir is just the opposite of current direction
            return Direction((curr_dir + 2) % len(Direction))
        return self._get_next_direction(action, curr_dir)

    @staticmethod
    def _get_next_direction(action: DAction, curr_dir: Direction) -> Direction:
        if action == TURN_RIGHT:
            return Direction((curr_dir + 1) % len(Direction))
        if action == TURN_LEFT:
            return Direction((curr_dir - 1) % len(Direction))
        return curr_dir

    @staticmethod
    def _get_next_speed(action: DAction, curr_speed: Speed) -> Speed:
        if action == DO_NOTHING:
            return curr_speed
        if action in (TURN_LEFT, TURN_RIGHT):
            if curr_speed == Speed.FORWARD_FAST:
                return Speed.FORWARD_SLOW
            return curr_speed
        if action == ACCELERATE:
            return Speed(min(curr_speed + 1, Speed.FORWARD_FAST))
        return Speed(max(curr_speed - 1, Speed.REVERSE))

    def _get_next_coord(
        self,
        curr_coord: Coord,
        speed: Speed,
        move_dir: Direction,
        vehicle_coords: Set[Coord],
    ) -> Tuple[Coord, CollisionType, Optional[Coord]]:
        # assumes curr_coord isn't in vehicle coords
        next_coord = curr_coord
        collision = CollisionType.NONE
        hit_vehicle_coord = None
        for i in range(abs(speed - Speed.STOPPED)):
            next_coord = self.grid.get_next_coord(
                curr_coord, move_dir, ignore_blocks=False
            )
            if next_coord == curr_coord:
                collision = CollisionType.OBSTACLE
                break
            elif next_coord in vehicle_coords:
                collision = CollisionType.VEHICLE
                hit_vehicle_coord = next_coord
                next_coord = curr_coord
                break
            curr_coord = next_coord

        return (next_coord, collision, hit_vehicle_coord)

    def _get_obs(self, state: DState) -> Dict[M.AgentID, DObs]:
        obs: Dict[M.AgentID, DObs] = {}
        for i in range(len(self.possible_agents)):
            local_cell_obs = self._get_local_cell__obs(
                i, [vs.coord for vs in state], state[i].facing_dir, state[i].dest_coord
            )
            obs[i] = (
                local_cell_obs,
                state[i].speed,
                state[i].dest_coord,
                state[i].dest_reached,
                state[i].crashed,
            )
        return obs

    def _get_local_cell__obs(
        self,
        agent_id: int,
        vehicle_coords: Sequence[Coord],
        facing_dir: Direction,
        dest_coord: Coord,
    ) -> Tuple[int, ...]:
        obs_depth = self._obs_front + self._obs_back + 1
        obs_width = (2 * self._obs_side) + 1
        agent_coord = vehicle_coords[agent_id]

        cell_obs = []
        for col, row in product(range(obs_width), range(obs_depth)):
            obs_grid_coord = self._map_obs_to_grid_coord(
                (col, row), agent_coord, facing_dir
            )
            if obs_grid_coord is None or obs_grid_coord in self.grid.block_coords:
                cell_obs.append(WALL)
            elif obs_grid_coord in vehicle_coords:
                cell_obs.append(VEHICLE)
            elif obs_grid_coord == dest_coord:
                cell_obs.append(DESTINATION)
            else:
                cell_obs.append(EMPTY)
        return tuple(cell_obs)

    def _map_obs_to_grid_coord(
        self, obs_coord: Coord, agent_coord: Coord, facing_dir: Direction
    ) -> Optional[Coord]:
        if facing_dir == Direction.NORTH:
            grid_row = agent_coord[1] + obs_coord[1] - self._obs_front
            grid_col = agent_coord[0] + obs_coord[0] - self._obs_side
        elif facing_dir == Direction.EAST:
            grid_row = agent_coord[1] + obs_coord[0] - self._obs_side
            grid_col = agent_coord[0] - obs_coord[1] + self._obs_front
        elif facing_dir == Direction.SOUTH:
            grid_row = agent_coord[1] - obs_coord[1] + self._obs_front
            grid_col = agent_coord[0] - obs_coord[0] + self._obs_side
        else:
            grid_row = agent_coord[1] - obs_coord[0] + self._obs_side
            grid_col = agent_coord[0] + obs_coord[1] - self._obs_front

        if 0 <= grid_row < self.grid.height and 0 <= grid_col < self.grid.width:
            return (grid_col, grid_row)
        return None

    def get_obs_coords(self, origin: Coord, facing_dir: Direction) -> List[Coord]:
        """Get the list of coords observed from agent at origin."""
        obs_depth = self._obs_front + self._obs_back + 1
        obs_width = (2 * self._obs_side) + 1
        obs_coords: List[Coord] = []
        for col, row in product(range(obs_width), range(obs_depth)):
            obs_grid_coord = self._map_obs_to_grid_coord((col, row), origin, facing_dir)
            if obs_grid_coord is not None:
                obs_coords.append(obs_grid_coord)
        return obs_coords

    def _get_rewards(
        self, state: DState, next_state: DState, collision_types: List[CollisionType]
    ) -> Dict[M.AgentID, SupportsFloat]:
        rewards: Dict[M.AgentID, SupportsFloat] = {}
        for i in range(len(self.possible_agents)):
            if state[i].crashed or state[i].dest_reached:
                # already in terminal/rewarded state
                r_i = 0.0
            elif (
                self._obstacle_collisions
                and collision_types[i] == CollisionType.OBSTACLE
            ):
                # Treat as if crashed into a vehicle
                r_i = self.R_CRASH_VEHICLE
            elif collision_types[i] == CollisionType.VEHICLE:
                r_i = self.R_CRASH_VEHICLE
            elif next_state[i].dest_reached:
                r_i = self.R_DESTINATION_REACHED
            else:
                r_i = self.R_STEP_COST

            progress = state[i].min_dest_dist - next_state[i].min_dest_dist
            r_i += max(0, progress) * self.R_PROGRESS

            if (
                not self._obstacle_collisions
                and collision_types[i] == CollisionType.OBSTACLE
            ):
                r_i += self.R_CRASH_OBJECT
            rewards[i] = r_i
        return rewards


class DrivingGrid(Grid):
    """A grid for the Driving Problem."""

    def __init__(
        self,
        grid_width: int,
        grid_height: int,
        block_coords: Set[Coord],
        start_coords: List[Set[Coord]],
        dest_coords: List[Set[Coord]],
    ):
        super().__init__(grid_width, grid_height, block_coords)
        assert len(start_coords) == len(dest_coords)
        self.start_coords = start_coords
        self.dest_coords = dest_coords

        self.shortest_paths = self.get_all_shortest_paths(set.union(*dest_coords))

    @property
    def supported_num_agents(self) -> int:
        """Get the number of agents supported by this grid."""
        return len(self.start_coords)

    def get_shortest_path_distance(self, coord: Coord, dest: Coord) -> int:
        """Get the shortest path distance from coord to destination."""
        return int(self.shortest_paths[dest][coord])

    def get_max_shortest_path_distance(self) -> int:
        """Get the longest shortest path distance to any destination."""
        return int(max([max(d.values()) for d in self.shortest_paths.values()]))

    def get_ascii_repr(
        self,
        vehicle_coords: List[Coord],
        vehicle_dirs: List[Direction],
        vehicle_dests: List[Coord],
    ) -> str:
        """Get ascii repr of grid."""
        grid_repr = []
        for row in range(self.height):
            row_repr = []
            for col in range(self.width):
                coord = (col, row)
                if coord in self.block_coords:
                    row_repr.append("#")
                elif coord in vehicle_dests:
                    row_repr.append("D")
                else:
                    row_repr.append(".")
            grid_repr.append(row_repr)

        for coord, direction in zip(vehicle_coords, vehicle_dirs):
            grid_repr[coord[0]][coord[1]] = DIRECTION_ASCII_REPR[direction]

        return "\n".join(list(list((" ".join(r) for r in grid_repr))))


def parse_grid_str(grid_str: str, supported_num_agents: int) -> DrivingGrid:
    """Parse a str representation of a grid.

    Notes on grid str representation:

    . = empty/unallocated cell
    # = a block
    0, 1, ..., 9 = starting location for agent with given index
    + = starting point for any agent
    a, b, ..., j = destination location for agent with given index
                   (a=0, b=1, ..., j=9)
    - = destination location for any agent

    Examples (" " quotes and newline chars ommited):

    1. A 3x3 grid with two agents, one block, and where each agent has a single
    starting location and a single destination location.

    a1.
    .#.
    .0.

    2. A 6x6 grid with 4 common start and destination locations and many
       blocks. This grid can support up to 4 agents.

    +.##+#
    ..+#.+
    #.###.
    #.....
    -..##.
    #-.-.-

    """
    row_strs = grid_str.splitlines()
    assert len(row_strs) > 1
    assert all(len(row) == len(row_strs[0]) for row in row_strs)
    assert len(row_strs[0]) > 1

    grid_height = len(row_strs)
    grid_width = len(row_strs[0])

    agent_start_chars = set(["+"] + [str(i) for i in range(10)])
    agent_dest_chars = set(["-"] + [c for c in "abcdefghij"])

    block_coords: Set[Coord] = set()
    shared_start_coords: Set[Coord] = set()
    agent_start_coords_map: Dict[int, Set[Coord]] = {}
    shared_dest_coords: Set[Coord] = set()
    agent_dest_coords_map: Dict[int, Set[Coord]] = {}
    for r, c in product(range(grid_height), range(grid_width)):
        coord = (c, r)
        char = row_strs[r][c]

        if char == "#":
            block_coords.add(coord)
        elif char in agent_start_chars:
            if char != "+":
                agent_id = int(char)
                if agent_id not in agent_start_coords_map:
                    agent_start_coords_map[agent_id] = set()
                agent_start_coords_map[agent_id].add(coord)
            else:
                shared_start_coords.add(coord)
        elif char in agent_dest_chars:
            if char != "-":
                agent_id = ord(char) - ord("a")
                if agent_id not in agent_dest_coords_map:
                    agent_dest_coords_map[agent_id] = set()
                agent_dest_coords_map[agent_id].add(coord)
            else:
                shared_dest_coords.add(coord)

    assert (
        len(shared_start_coords) + len(agent_start_coords_map) >= supported_num_agents
    )
    assert len(shared_dest_coords) + len(agent_dest_coords_map) >= supported_num_agents

    included_agent_ids = list(set([*agent_start_coords_map, *agent_dest_coords_map]))
    if len(included_agent_ids) > 0:
        assert max(included_agent_ids) < supported_num_agents

    start_coords: List[Set[Coord]] = []
    dest_coords: List[Set[Coord]] = []
    for i in range(supported_num_agents):
        agent_start_coords = set(shared_start_coords)
        agent_start_coords.update(agent_start_coords_map.get(i, {}))
        start_coords.append(agent_start_coords)

        agent_dest_coords = set(shared_dest_coords)
        agent_dest_coords.update(agent_dest_coords_map.get(i, {}))
        dest_coords.append(agent_dest_coords)

    return DrivingGrid(
        grid_width=grid_width,
        grid_height=grid_height,
        block_coords=block_coords,
        start_coords=start_coords,
        dest_coords=dest_coords,
    )


def get_3x3_grid() -> DrivingGrid:
    """Generate a simple Driving 3-by-3 grid layout."""
    grid_str = "a1.\n" ".#.\n" ".0b\n"
    return parse_grid_str(grid_str, 2)


def get_4x4_intersection_grid() -> DrivingGrid:
    """Generate a 4-by-4 intersection grid layout."""
    grid_str = "#0b#\n" "d..3\n" "2..c\n" "#a1#\n"
    return parse_grid_str(grid_str, 4)


def get_6x6_intersection_grid() -> DrivingGrid:
    """Generate a 6-by-6 intersection grid layout."""
    grid_str = "##0b##\n" "##..##\n" "d....3\n" "2....c\n" "##..##\n" "##a1##\n"
    return parse_grid_str(grid_str, 4)


def get_7x7_crisscross_grid() -> DrivingGrid:
    """Generate a 7-by-7 Criss-Cross grid layout."""
    grid_str = (
        "#-#-#-#\n"
        "-.....+\n"
        "#.#.#.#\n"
        "-.....+\n"
        "#.#.#.#\n"
        "-.....+\n"
        "#+#+#+#\n"
    )
    return parse_grid_str(grid_str, 6)


def get_7x7_blocks_grid() -> DrivingGrid:
    """Generate a 7-by-7 blocks grid layout."""
    grid_str = (
        "#-...-#\n"
        "-##.##+\n"
        ".##.##.\n"
        ".......\n"
        ".##.##.\n"
        "-##.##+\n"
        "#+...+#\n"
    )
    return parse_grid_str(grid_str, 4)


def get_7x7_roundabout_grid() -> DrivingGrid:
    """Generate a 7-by-7 round-about grid layout."""
    grid_str = (
        "#-...-#\n"
        "-##.##+\n"
        ".#...#.\n"
        "...#...\n"
        ".#...#.\n"
        "-##.##+\n"
        "#+...+#\n"
    )
    return parse_grid_str(grid_str, 4)


def get_14x14_crisscross_grid() -> DrivingGrid:
    """Generate a 14-by-14 Criss-Cross grid layout."""
    grid_str = (
        "##-##-##-##-##\n"
        "##.##.##.##.##\n"
        "-............+\n"
        "##.##.##.##.##\n"
        "##.##.##.##.##\n"
        "-............+\n"
        "##.##.##.##.##\n"
        "##.##.##.##.##\n"
        "-............+\n"
        "##.##.##.##.##\n"
        "##.##.##.##.##\n"
        "-............+\n"
        "##.##.##.##.##\n"
        "##+##+##+##+##\n"
    )
    return parse_grid_str(grid_str, 4)


def get_14x14_blocks_grid() -> DrivingGrid:
    """Generate a 14-by-14 Blocks grid layout."""
    grid_str = (
        "#-..........-#\n"
        "-###.####.###+\n"
        ".###.####.###.\n"
        ".###.####.###.\n"
        "..............\n"
        ".###.####.###.\n"
        ".###.####.###.\n"
        ".###.####.###.\n"
        ".###.####.###.\n"
        "..............\n"
        ".###.####.###.\n"
        ".###.####.###.\n"
        "-###.####.###+\n"
        "#+..........+#\n"
    )
    return parse_grid_str(grid_str, 4)


def get_14x14_roundabout_wide_grid() -> DrivingGrid:
    """Generate a 14-by-14 Round About grid layout with wide roads."""
    grid_str = (
        "#-..........-#\n"
        "-#####..#####+\n"
        ".#####..#####.\n"
        ".#####..#####.\n"
        ".###......###.\n"
        ".###......###.\n"
        "......##......\n"
        "......##......\n"
        ".###......###.\n"
        ".###......###.\n"
        ".#####..#####.\n"
        ".#####..#####.\n"
        "-#####..#####+\n"
        "#+..........+#\n"
    )
    return parse_grid_str(grid_str, 4)


class DrivingGridGenerator(GridGenerator):
    """Class for generating grid layouts for Driving Environment.

    Generates starting and destination coords in an alternating pattern along
    the outside of edge of the grid.
    """

    def __init__(
        self,
        width: int,
        height: int,
        max_obstacle_size: int,
        max_num_obstacles: int,
        seed: Optional[int] = None,
    ):
        super().__init__(
            width,
            height,
            self._generate_mask(width, height),
            max_obstacle_size,
            max_num_obstacles,
            ensure_grid_connected=True,
            seed=seed,
        )

        self._start_coords = [self.mask for _ in range(len(self.mask))]
        self._dest_coords = [self.mask for _ in range(len(self.mask))]

    def _generate_mask(self, width: int, height: int) -> Set[Coord]:
        start = 1
        mask = set()
        for x in range(start, width, 2):
            mask.add((x, 0))
            mask.add((x, height - 1))

        for y in range(start, height, 2):
            mask.add((0, y))
            mask.add((width - 1, y))

        return mask

    def generate(self) -> DrivingGrid:
        """Generate a new Driving grid."""
        base_grid = super().generate()
        driving_grid = DrivingGrid(
            base_grid.width,
            base_grid.height,
            base_grid.block_coords,
            self._start_coords,
            self._dest_coords,
        )
        return driving_grid


# (generator params, finite horizon step limit)
SUPPORTED_GEN_PARAMS = {
    "7x7": (
        {
            "width": 7,
            "height": 7,
            "max_obstacle_size": 2,
            "max_num_obstacles": 21,  # size * 3
        },
        40,
    ),
    "14x14": (
        {
            "width": 14,
            "height": 14,
            "max_obstacle_size": 3,
            "max_num_obstacles": 42,  # size * 3
        },
        100,
    ),
    "28x28": (
        {
            "width": 28,
            "height": 28,
            "max_obstacle_size": 4,
            "max_num_obstacles": 84,  # size * 3
        },
        200,
    ),
}


#  (grid_make_fn, max step_limit, )
SUPPORTED_GRIDS = {
    "3x3": (get_3x3_grid, 30),
    "4x4Intersection": (get_4x4_intersection_grid, 20),
    "6x6Intersection": (get_6x6_intersection_grid, 20),
    "7x7CrissCross": (get_7x7_crisscross_grid, 50),
    "7x7Blocks": (get_7x7_blocks_grid, 50),
    "7x7RoundAbout": (get_7x7_roundabout_grid, 50),
    "14x14Blocks": (get_14x14_blocks_grid, 100),
    "14x14CrissCross": (get_14x14_crisscross_grid, 100),
    "14x14WideRoundAbout": (get_14x14_roundabout_wide_grid, 50),
}


def load_grid(grid_name: str) -> DrivingGrid:
    """Load grid with given name."""
    grid_name = grid_name
    assert grid_name in SUPPORTED_GRIDS, (
        f"Unsupported grid name '{grid_name}'. Grid name must be one of: "
        f"{SUPPORTED_GRIDS.keys()}."
    )
    return SUPPORTED_GRIDS[grid_name][0]()



"""Class for displaying renders of grid-world environments.

The Viewer is seperated from the rendering functionality, in case users want
to use the renders without displaying them, and to seperate out dependencies
in that case.

This implemention is based on the gym-minigrid library:
- github.com/maximecb/gym-minigrid/blob/master/gym_minigrid/minigrid.py
"""
import sys
from typing import Tuple, Optional

import numpy as np

try:
    import matplotlib.pyplot as plt
    import matplotlib.gridspec as gridspec
except ImportError:
    print('To display the environment in a window, please install matplotlib.')
    sys.exit(-1)


class GWViewer:
    """Handles displaying grid world renders."""

    def __init__(self,
                 title: str,
                 figsize: Tuple[int, int] = (9, 9),
                 num_agent_displays: Optional[int] = None):
        self._title = title
        self._figsize = figsize
        self._num_agent_displays = num_agent_displays

        self._fig = plt.figure(
            num=title, figsize=figsize, constrained_layout=True
        )
        self._bg = self._fig.canvas.copy_from_bbox(self._fig.bbox)

        if num_agent_displays is None:
            spec = gridspec.GridSpec(1, 1, figure=self._fig)
        else:
            spec = gridspec.GridSpec(num_agent_displays, 2, figure=self._fig)

        # Main world view is entire first column
        self._main_ax = self._fig.add_subplot(spec[:, 0])
        self._main_imshow_obj = None

        self._agent_axs = []
        self._agent_imshow_objs = []              # type: ignore
        if num_agent_displays is not None:
            for i in range(num_agent_displays):
                ax = self._fig.add_subplot(spec[i, 1])
                self._agent_axs.append(ax)
                self._agent_imshow_objs.append(None)

        for ax in [self._main_ax] + self._agent_axs:
            # Turn off x/y axis numbering/ticks
            ax.xaxis.set_ticks_position('none')
            ax.yaxis.set_ticks_position('none')
            ax.set_xticklabels([])
            ax.set_yticklabels([])

        # Flag indicating window was closed
        self._closed = False

        # pylint: disable=[unused-argument]
        def close_handler(event):
            self._closed = True

        self._fig.canvas.mpl_connect('close_event', close_handler)

    def update_img(self, img: np.ndarray, agent_idx: Optional[int] = None):
        """Update image to be shown for env or agent."""
        if agent_idx is None:
            ax = self._main_ax
            imshow_obj = self._main_imshow_obj
        else:
            ax = self._agent_axs[agent_idx]
            imshow_obj = self._agent_imshow_objs[agent_idx]

        # Show the first image of the environment
        if imshow_obj is None:
            imshow_obj = ax.imshow(
                img, interpolation='bilinear', origin='upper', animated=True
            )

            if agent_idx is None:
                self._main_imshow_obj = imshow_obj
            else:
                self._agent_imshow_objs[agent_idx] = imshow_obj

        imshow_obj.set_data(img)

    def display_img(self):
        """Show current image/s."""
        self._fig.canvas.restore_region(self._bg)
        # Request the window be redrawn
        self._fig.canvas.blit(self._fig.bbox)
        self._fig.canvas.flush_events()

        # Let matplotlib process UI events
        # This is needed for interactive mode to work properly
        plt.pause(0.01)

    @staticmethod
    def show(block: bool = True) -> None:
        """Show the window, and start an event loop."""
        # If not blocking, trigger interactive mode
        if not block:
            plt.ion()

        # Show the plot
        # In non-interative mode, this enters the matplotlib event loop
        # In interactive mode, this call does not block
        plt.show()

    def close(self) -> None:
        """Close the window."""
        plt.close(self._fig)
        self._closed = True



"""Functions and classes for rendering grid world environments.

This implemention is based on the gym-minigrid library:
- github.com/maximecb/gym-minigrid/blob/master/gym_minigrid/minigrid.py
"""
import enum
import math
from itertools import product
from typing import Callable, Dict, List, Optional, Set, Tuple, Union

import numpy as np

from posggym.envs.grid_world.core import Coord, Direction, Grid


# Size in pixels of a tile in the full-scale human view
TILE_PIXELS = 32

# Map of color names to RGB values
COLORS = {
    "red": np.array([255, 0, 0]),
    "blue": np.array([0, 0, 255]),
    "green": np.array([0, 255, 0]),
    "purple": np.array([112, 39, 195]),
    "yellow": np.array([255, 255, 0]),
    "grey": np.array([100, 100, 100]),
    "white": np.array([255, 255, 255]),
    "cyan": np.array([0, 195, 195]),
}

COLOR_TO_IDX = {
    "red": 0,
    "blue": 1,
    "green": 2,
    "purple": 3,
    "yellow": 4,
    "grey": 5,
    "white": 6,
    "cyan": 7,
}

AGENT_COLORS = ["red", "blue", "green", "purple", "yellow", "white", "cyan"]


def get_agent_color(agent_id: int) -> str:
    """Get color for agent."""
    return AGENT_COLORS[agent_id]


# Map of agent direction indices to vectors
# Ordering matches ordering in Direction enum
DIR_TO_VEC = [
    # NORTH (negative Y)
    np.array((0, -1)),
    # EAST (positive X)
    np.array((1, 0)),
    # SOUTH (positive Y)
    np.array((0, 1)),
    # WEST (negative X)
    np.array((-1, 0)),
]

DIR_TO_THETA = [
    # NORTH
    0.5 * math.pi * 2,
    # EAST
    0.5 * math.pi * 1,
    # SOUTH
    0.5 * math.pi * 0,
    # WEST
    0.5 * math.pi * 3,
]


class Shape(enum.Enum):
    """An Object shape."""

    RECTANGLE = enum.auto()
    TRIANGLE = enum.auto()
    CIRCLE = enum.auto()
    LINE = enum.auto()


class GWObject:
    """An object in the grid world."""

    def __init__(
        self,
        coord: Coord,
        color: str,
        shape: Shape,
        direction: Optional[Direction] = None,
        alpha: Optional[float] = None,
    ):
        if shape == Shape.TRIANGLE:
            assert direction is not None

        self.coord = coord
        self.color = color
        self.shape = shape
        self.direction = direction
        self.alpha = alpha

    def render(self, img: np.ndarray):
        """Draw object into image cell."""
        if self.shape == Shape.RECTANGLE:
            _fill_coords(img, _point_in_rect(0, 1, 0, 1), COLORS[self.color])
        elif self.shape == Shape.CIRCLE:
            _fill_coords(img, _point_in_circle(0.5, 0.5, 0.31), COLORS[self.color])
        elif self.shape == Shape.TRIANGLE:
            tri_fn = _point_in_triangle((0.12, 0.19), (0.87, 0.50), (0.12, 0.81))
            theta = DIR_TO_THETA[self.direction]  # type: ignore
            tri_fn = _rotate_fn(tri_fn, cx=0.5, cy=0.5, theta=theta)
            _fill_coords(img, tri_fn, COLORS[self.color])
        else:
            raise AttributeError("Unsupported Object shape: {self.shape}")

        if self.alpha is not None:
            _highlight_img(img, alpha=self.alpha)


class GWRenderer:
    """Handles generating grid world renders."""

    def __init__(
        self,
        n_agents: int,
        grid: Grid,
        static_objs: List[GWObject],
        render_blocks: bool = True,
    ):
        self._n_agents = n_agents
        self._grid = grid
        self._static_objs = static_objs
        self._width, self._height = grid.width, grid.height

        self._tiles: List[List[List[Optional[GWObject]]]] = []
        for x in range(self._width):
            self._tiles.append([])
            for y in range(self._height):
                self._tiles[x].append([])
                self._tiles[x][y].append(None)

        if render_blocks:
            for block_coord in grid.block_coords:
                block_obj = GWObject(block_coord, "grey", Shape.RECTANGLE)
                self._tiles[block_coord[0]][block_coord[1]].append(block_obj)

        for obj in static_objs:
            self._tiles[obj.coord[0]][obj.coord[1]].append(obj)

    def _render_tile(
        self, coords: Coord, other_objs: List[GWObject], observed: bool, tile_size: int
    ) -> np.ndarray:
        img = np.zeros(shape=(tile_size, tile_size, 3), dtype=np.uint8)

        # Draw the grid lines (tope and left edges)
        _fill_coords(img, _point_in_rect(0, 0.031, 0, 1), COLORS["grey"])
        _fill_coords(img, _point_in_rect(0, 1, 0, 0.031), COLORS["grey"])

        for obj in self._tiles[coords[0]][coords[1]]:
            if obj is not None:
                obj.render(img)

        for obj in other_objs:
            obj.render(img)

        if observed:
            _highlight_img(img)

        return img

    def _render_obj_tile(
        self, obj: GWObject, observed: bool, tile_size: int
    ) -> np.ndarray:
        img = np.zeros(shape=(tile_size, tile_size, 3), dtype=np.uint8)

        # Draw the grid lines (tope and left edges)
        _fill_coords(img, _point_in_rect(0, 0.031, 0, 1), COLORS["grey"])
        _fill_coords(img, _point_in_rect(0, 1, 0, 0.031), COLORS["grey"])

        obj.render(img)

        if observed:
            _highlight_img(img)

        return img

    def render(
        self,
        agent_coords: Tuple[Coord, ...],
        agent_obs_coords: Optional[Tuple[List[Coord], ...]],
        agent_dirs: Optional[Tuple[Direction, ...]],
        other_objs: Optional[List[GWObject]],
        agent_colors: Optional[Tuple[str, ...]] = None,
        tile_size: int = TILE_PIXELS,
    ) -> np.ndarray:
        """Generate Grid-World render.

        Returns a numpy array image representation of the grid world. This
        includes the agents inside the grid-world.
        """
        width_px = self._width * tile_size
        height_px = self._height * tile_size

        img = np.zeros(shape=(width_px, height_px, 3), dtype=np.uint8)

        observed_coords: Set[Coord] = set()
        if agent_obs_coords is not None:
            observed_coords.update(*agent_obs_coords)

        dynamic_objs_map: Dict[Coord, List[GWObject]] = {}
        if other_objs is None:
            other_objs = []

        for obj in other_objs:
            if obj.coord not in dynamic_objs_map:
                dynamic_objs_map[obj.coord] = []
            dynamic_objs_map[obj.coord].append(obj)

        if agent_dirs is None:
            agent_dirs = tuple(Direction.NORTH for _ in range(self._n_agents))

        if agent_colors is None:
            assert self._n_agents <= len(AGENT_COLORS), (
                "Agent colors must be specified when rendering envs with "
                f"more than {len(AGENT_COLORS)-1} agents."
            )
            agent_colors = tuple(AGENT_COLORS[i] for i in range(self._n_agents))

        for coord, direction, color in zip(agent_coords, agent_dirs, agent_colors):
            if coord not in dynamic_objs_map:
                dynamic_objs_map[coord] = []
            agent_obj = GWObject(coord, color, Shape.TRIANGLE, direction)
            dynamic_objs_map[coord].append(agent_obj)

        for x in range(self._width):
            for y in range(self._height):
                coord = (x, y)
                tile_img = self._render_tile(
                    coord,
                    dynamic_objs_map.get(coord, []),
                    coord in observed_coords,
                    tile_size,
                )

                xmin, xmax = x * tile_size, (x + 1) * tile_size
                ymin, ymax = y * tile_size, (y + 1) * tile_size
                img[xmin:xmax, ymin:ymax, :] = tile_img

        img = img.transpose(1, 0, 2)
        return img

    def render_all_agent_obs(
        self,
        env_img: np.ndarray,
        agent_coords: Tuple[Coord, ...],
        agent_dirs: Union[Direction, Tuple[Direction, ...]],
        agent_obs_dims: Union[Tuple[int, int, int], Tuple[Tuple[int, int, int], ...]],
        out_of_bounds_obj: GWObject,
        agent_obs_coords: Optional[Tuple[List[Coord], ...]] = None,
        tile_size: int = TILE_PIXELS,
    ) -> Tuple[np.ndarray, ...]:
        """Generate seperate render for a each grid-world agent.

        Returns a numpy array image representation of the agents observation
        of the grid world for each agent.

        Assumes env_img was render generated by the Renderer.render function.
        """
        agent_obs_imgs = []
        for i in range(self._n_agents):
            if isinstance(agent_dirs, tuple):
                agent_dir = agent_dirs[i]
            else:
                agent_dir = agent_dirs

            if isinstance(agent_obs_dims[0], tuple):
                obs_dims = agent_obs_dims[i]
            else:
                obs_dims = agent_obs_dims  # type: ignore

            if isinstance(agent_obs_coords, tuple):
                obs_coords = agent_obs_coords[i]
            else:
                obs_coords = None  # type: ignore

            agent_img = self.render_agent_obs(
                env_img,
                agent_coords[i],
                agent_dir,
                obs_dims,  # type: ignore
                out_of_bounds_obj,
                obs_coords,
                tile_size,
            )
            agent_obs_imgs.append(agent_img)
        return tuple(agent_obs_imgs)

    def render_agent_obs(
        self,
        env_img: np.ndarray,
        agent_coord: Coord,
        agent_dir: Direction,
        agent_obs_dims: Tuple[int, int, int],
        out_of_bounds_obj: GWObject,
        agent_obs_coords: Optional[List[Coord]] = None,
        tile_size: int = TILE_PIXELS,
    ) -> np.ndarray:
        """Generate render for a single grid-world agent.

        Arguments
        ---------
        env_img : np.ndarray
            image render of entire environment assumed to be generated by the
            Renderer.render function.
        agent_coord: Coord
            Coordinates of agent within the environment grid
        agent_obs_dims: (int, int, int)
            Dimensions of agents observation in terms of number of cells
            in front, behind and to each side.
        agent_dir: Direction
            Direction the agent is facing within the environment grid.
        out_of_bounds_obj: GWObject
            Render object to use for cells in observation that are outside of
            environment grid (e.g. when agent is at the boundary of the
            environment)
        agent_obs_coords: Optional[List[Coord]]
            Optional list of environment coordinates for cells observed by
            agent. If provided, then treats all unobserved cells as out of
            bounds. If None, then assumes all coordinates in agents observation
            dims are observed.
        tile_size: int
            Defines the resolution of the image.

        Returns a numpy array image representation of the agents observation
        of the grid world.

        """
        # Reverse transpose to handle coord system properly
        env_img = env_img.transpose(1, 0, 2)
        env_width = env_img.shape[0] / tile_size
        env_height = env_img.shape[1] / tile_size

        obs_x_dim = (2 * agent_obs_dims[2]) + 1
        obs_y_dim = agent_obs_dims[0] + agent_obs_dims[1] + 1

        obs_width_px = obs_x_dim * tile_size
        obs_height_px = obs_y_dim * tile_size
        obs_img = np.zeros(shape=(obs_width_px, obs_height_px, 3), dtype=np.uint8)

        for (obs_x, obs_y) in product(range(obs_x_dim), range(obs_y_dim)):
            # Need to map to env (x, y)
            env_x, env_y = self._map_obs_to_env_coord(
                (obs_x, obs_y),
                agent_coord,
                agent_dir,
                agent_obs_forward_dim=agent_obs_dims[0],
                agent_obs_side_dim=agent_obs_dims[2],
            )

            cell_out_of_bounds = (
                not 0 <= env_x < env_width or not 0 <= env_y < env_height
            )
            cell_observed = (
                agent_obs_coords is None or (env_x, env_y) in agent_obs_coords
            )

            if cell_out_of_bounds or not cell_observed:
                tile_img = self._render_obj_tile(out_of_bounds_obj, False, tile_size)
            else:
                env_xmin, env_xmax = env_x * tile_size, (env_x + 1) * tile_size
                env_ymin, env_ymax = env_y * tile_size, (env_y + 1) * tile_size
                tile_img = env_img[env_xmin:env_xmax, env_ymin:env_ymax, :]

                # rotate tile so its oriented to agent's viewpoint
                # function rotates counter-clockwise
                k = (len(Direction) - agent_dir) % len(Direction)
                tile_img = np.rot90(tile_img, k)

            obs_xmin, obs_xmax = obs_x * tile_size, (obs_x + 1) * tile_size
            obs_ymin, obs_ymax = obs_y * tile_size, (obs_y + 1) * tile_size
            obs_img[obs_xmin:obs_xmax, obs_ymin:obs_ymax, :] = tile_img

        # Needed to swap to formate expected by matplotlib.imshow function
        obs_img = obs_img.transpose(1, 0, 2)
        return obs_img

    def _map_obs_to_env_coord(
        self,
        obs_coord: Coord,
        agent_coord: Coord,
        facing_dir: Direction,
        agent_obs_forward_dim: int,
        agent_obs_side_dim: int,
    ) -> Coord:
        grid_row = agent_coord[1]
        grid_col = agent_coord[0]

        if facing_dir == Direction.NORTH:
            grid_row += obs_coord[1] - agent_obs_forward_dim
            grid_col += obs_coord[0] - agent_obs_side_dim
        elif facing_dir == Direction.EAST:
            grid_row += obs_coord[0] - agent_obs_side_dim
            grid_col += -obs_coord[1] + agent_obs_forward_dim
        elif facing_dir == Direction.SOUTH:
            grid_row += -obs_coord[1] + agent_obs_forward_dim
            grid_col += -obs_coord[0] + agent_obs_side_dim
        else:
            grid_row += -obs_coord[0] + agent_obs_side_dim
            grid_col += obs_coord[1] - agent_obs_forward_dim
        return (grid_col, grid_row)


def _fill_coords(img: np.ndarray, fn: Callable, color: np.ndarray) -> np.ndarray:
    """Fill pixels of an image with coordinates matching a filter function."""
    for y in range(img.shape[0]):
        for x in range(img.shape[1]):
            yf = (y + 0.5) / img.shape[0]
            xf = (x + 0.5) / img.shape[1]
            if fn(xf, yf):
                img[y, x] = color

    return img


def _rotate_fn(fin: Callable, cx: float, cy: float, theta: float) -> Callable:
    def _fout(x, y):
        x = x - cx
        y = y - cy

        x2 = cx + x * math.cos(-theta) - y * math.sin(-theta)
        y2 = cy + y * math.cos(-theta) + x * math.sin(-theta)

        return fin(x2, y2)

    return _fout


def _point_in_line(x0: float, y0: float, x1: float, y1: float, r: float) -> Callable:
    p0 = np.array([x0, y0])
    p1 = np.array([x1, y1])
    direction = p1 - p0
    dist = np.linalg.norm(direction)
    direction = direction / dist

    xmin = min(x0, x1) - r
    xmax = max(x0, x1) + r
    ymin = min(y0, y1) - r
    ymax = max(y0, y1) + r

    def _fn(x, y):
        # Fast, early escape test
        if x < xmin or x > xmax or y < ymin or y > ymax:
            return False

        q = np.array([x, y])
        pq = q - p0

        # Closest point on line
        a = np.dot(pq, direction)
        a = np.clip(a, 0, dist)
        p = p0 + a * direction

        dist_to_line = np.linalg.norm(q - p)
        return dist_to_line <= r

    return _fn


def _point_in_circle(cx, cy, r):
    def _fn(x, y):
        return (x - cx) * (x - cx) + (y - cy) * (y - cy) <= r * r

    return _fn


def _point_in_rect(xmin: float, xmax: float, ymin: float, ymax: float) -> Callable:
    def _fn(x, y):
        return xmin <= x <= xmax and ymin <= y <= ymax

    return _fn


def _point_in_triangle(a, b, c):
    a = np.array(a)
    b = np.array(b)
    c = np.array(c)

    def _fn(x, y):
        v0 = c - a
        v1 = b - a
        v2 = np.array((x, y)) - a

        # Compute dot products
        dot00 = np.dot(v0, v0)
        dot01 = np.dot(v0, v1)
        dot02 = np.dot(v0, v2)
        dot11 = np.dot(v1, v1)
        dot12 = np.dot(v1, v2)

        # Compute barycentric coordinates
        inv_denom = 1 / (dot00 * dot11 - dot01 * dot01)
        u = (dot11 * dot02 - dot01 * dot12) * inv_denom
        v = (dot00 * dot12 - dot01 * dot02) * inv_denom

        # Check if point is in triangle
        return (u >= 0) and (v >= 0) and (u + v) < 1

    return _fn


def _highlight_img(img: np.ndarray, color=(255, 255, 255), alpha=0.2):
    """Add highlighting to an image."""
    blend_img = img + alpha * (np.array(color, dtype=np.uint8) - img)
    blend_img = blend_img.clip(0, 255).astype(np.uint8)
    img[:, :, :] = blend_img
